# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Kernel manager that optionally seeds kernel memory."""
import os
from typing import List, Optional

from jupyter_client.ioloop import AsyncIOLoopKernelManager
from jupyter_server.services.kernels.kernelmanager import AsyncMappingKernelManager
from traitlets import default


class SeedingMappingKernelManager(AsyncMappingKernelManager):
    """Extends the server's kernel manager to optionally execute the contents
    of a notebook on a kernel when it starts.
    """

    _seed_source: Optional[List]
    _seed_kernelspec: Optional[str]

    @default("root_dir")
    def _default_root_dir(self):
        return os.getcwd()

    def _kernel_manager_class_default(self):
        return "kernel_gateway.services.kernels.manager.KernelGatewayIOLoopKernelManager"

    @property
    def seed_kernelspec(self) -> Optional[str]:
        """Gets the kernel spec name for run the seed notebook.

        Prefers the spec name forced by configuration over the spec in the
        seed notebook itself.

        Returns
        -------
        str
            Name of the notebook kernelspec or None if no seed notebook exists
        """
        if hasattr(self, "_seed_kernelspec"):
            return self._seed_kernelspec

        if self.parent.seed_notebook:
            if self.parent.force_kernel_name:
                self._seed_kernelspec = self.parent.force_kernel_name
            else:
                self._seed_kernelspec = self.parent.seed_notebook["metadata"]["kernelspec"]["name"]
        else:
            self._seed_kernelspec = None

        return self._seed_kernelspec

    @property
    def seed_source(self) -> Optional[List]:
        """Gets the source of the seed notebook in cell order.

        Returns
        -------
        list
            Notebook code cell contents or None if no seed notebook exists
        """
        if hasattr(self, "_seed_source"):
            return self._seed_source

        if self.parent.seed_notebook:
            self._seed_source = [
                cell["source"]
                for cell in self.parent.seed_notebook.cells
                if cell["cell_type"] == "code"
            ]
        else:
            self._seed_source = None

        return self._seed_source

    async def start_seeded_kernel(self, *args, **kwargs):
        """Start a kernel using the language specified in the seed notebook.

        Run synchronously so that any exceptions thrown while seed rise up
        to the caller.
        """
        kwargs["kernel_name"] = self.seed_kernelspec
        await self.start_kernel(*args, **kwargs)

    async def start_kernel(self, *args, **kwargs):
        """Starts a kernel and then executes a list of code cells on it if a
        seed notebook exists.
        """
        if self.parent.force_kernel_name:
            kwargs["kernel_name"] = self.parent.force_kernel_name
        kernel_id = await super().start_kernel(*args, **kwargs)

        if kernel_id and self.seed_source is not None:
            # Only run source if the kernel spec matches the notebook kernel spec
            kernel = self.get_kernel(kernel_id)
            if kernel.kernel_name == self.seed_kernelspec:
                # Create a client to talk to the kernel
                client = kernel.client()
                # Clone client session. Workaround duplicate signatures due to shared digest_history
                # This shouldn't be necessary after upstream fixes.
                client.session = type(client.session)(
                    config=kernel.session.config,
                    key=kernel.session.key,
                )
                # Only start channels and wait for ready in HTTP mode
                client.start_channels()
                await client.wait_for_ready()
                for code in self.seed_source:
                    # Check with the personality whether it wants the cell
                    # executed
                    if self.parent.personality.should_seed_cell(code):
                        client.execute(code)
                        msg_type = "kernel_info_reply"
                        while msg_type == "kernel_info_reply":
                            msg = await client.get_shell_msg()
                            msg_type = msg["msg_type"]
                            if msg["content"]["status"] != "ok":
                                # Shutdown the channels to remove any lingering ZMQ messages
                                client.stop_channels()
                                # Shutdown the kernel
                                await self.shutdown_kernel(kernel_id)
                                raise RuntimeError("Error seeding kernel memory", msg["content"])
                # Shutdown the channels to remove any lingering ZMQ messages
                client.stop_channels()
        return kernel_id


class KernelGatewayIOLoopKernelManager(AsyncIOLoopKernelManager):
    """Extends the IOLoopKernelManager used by the SeedingMappingKernelManager.

    Sets the environment variable 'KERNEL_GATEWAY' to '1' to indicate that the
    kernel is executing within a Jupyter Kernel Gateway instance. Removes the
    KG_AUTH_TOKEN from the environment variables passed to the kernel when it
    starts.
    """

    async def _async_launch_kernel(self, kernel_cmd, **kw):
        # TODO - should probably figure out a better place to deal with this
        env = kw["env"]
        env["KERNEL_GATEWAY"] = "1"
        if "KG_AUTH_TOKEN" in env:
            del env["KG_AUTH_TOKEN"]
        return await super()._async_launch_kernel(kernel_cmd, **kw)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tornado handlers for the base of the API."""

import jupyter_server.base.handlers as server_handlers
from tornado import web

from ..mixins import CORSMixin, JSONErrorsMixin, TokenAuthorizationMixin


class APIVersionHandler(
    TokenAuthorizationMixin, CORSMixin, JSONErrorsMixin, server_handlers.APIVersionHandler
):
    """Extends the notebook server base API handler with token auth, CORS, and
    JSON errors.
    """


class NotFoundHandler(JSONErrorsMixin, web.RequestHandler):
    """Catches all requests and responds with 404 JSON messages.

    Installed as the fallback error for all unhandled requests.

    Raises
    ------
    tornado.web.HTTPError
        Always 404 Not Found
    """

    def prepare(self):
        raise web.HTTPError(404)


default_handlers = [(r"/api", APIVersionHandler), (r"/(.*)", NotFoundHandler)]



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tests for handler mixins."""

import json
from unittest.mock import Mock

import pytest
from tornado import web

from kernel_gateway.mixins import JSONErrorsMixin, TokenAuthorizationMixin


class SuperTokenAuthHandler:
    """Super class for the handler using TokenAuthorizationMixin."""

    is_prepared = False

    def prepare(self):
        # called by the mixin when authentication succeeds
        self.is_prepared = True


class CustomTokenAuthHandler(TokenAuthorizationMixin, SuperTokenAuthHandler):
    """Implementation that uses the TokenAuthorizationMixin for testing."""

    def __init__(self, token=""):
        self.settings = {"kg_auth_token": token}
        self.arguments = {}
        self.response = None
        self.status_code = None

    def send_error(self, status_code):
        self.status_code = status_code

    def get_argument(self, name, default=""):
        return self.arguments.get(name, default)


@pytest.fixture()
def auth_mixin():
    auth_mixin_instance = CustomTokenAuthHandler("YouKnowMe")
    return auth_mixin_instance


class TestTokenAuthMixin:
    """Unit tests the Token authorization mixin."""

    def test_no_token_required(self, auth_mixin):
        """Status should be None."""
        auth_mixin.request = Mock({})
        auth_mixin.settings["kg_auth_token"] = ""
        auth_mixin.prepare()
        assert auth_mixin.is_prepared is True
        assert auth_mixin.status_code is None

    def test_missing_token(self, auth_mixin):
        """Status should be 'unauthorized'."""
        attrs = {"headers": {}}
        auth_mixin.request = Mock(**attrs)
        auth_mixin.prepare()
        assert auth_mixin.is_prepared is False
        assert auth_mixin.status_code == 401

    def test_valid_header_token(self, auth_mixin):
        """Status should be None."""
        attrs = {"headers": {"Authorization": "token YouKnowMe"}}
        auth_mixin.request = Mock(**attrs)
        auth_mixin.prepare()
        assert auth_mixin.is_prepared is True
        assert auth_mixin.status_code is None

    def test_wrong_header_token(self, auth_mixin):
        """Status should be 'unauthorized'."""
        attrs = {"headers": {"Authorization": "token NeverHeardOf"}}
        auth_mixin.request = Mock(**attrs)
        auth_mixin.prepare()
        assert auth_mixin.is_prepared is False
        assert auth_mixin.status_code == 401

    def test_valid_url_token(self, auth_mixin):
        """Status should be None."""
        auth_mixin.arguments["token"] = "YouKnowMe"
        attrs = {"headers": {}}
        auth_mixin.request = Mock(**attrs)
        auth_mixin.prepare()
        assert auth_mixin.is_prepared is True
        assert auth_mixin.status_code is None

    def test_wrong_url_token(self, auth_mixin):
        """Status should be 'unauthorized'."""
        auth_mixin.arguments["token"] = "NeverHeardOf"
        attrs = {"headers": {}}
        auth_mixin.request = Mock(**attrs)
        auth_mixin.prepare()
        assert auth_mixin.is_prepared is False
        assert auth_mixin.status_code == 401

    def test_differing_tokens_valid_url(self, auth_mixin):
        """Status should be None, URL token takes precedence"""
        auth_mixin.arguments["token"] = "YouKnowMe"
        attrs = {"headers": {"Authorization": "token NeverHeardOf"}}
        auth_mixin.request = Mock(**attrs)
        auth_mixin.prepare()
        assert auth_mixin.is_prepared is True
        assert auth_mixin.status_code is None

    def test_differing_tokens_wrong_url(self, auth_mixin):
        """Status should be 'unauthorized', URL token takes precedence"""
        attrs = {"headers": {"Authorization": "token YouKnowMe"}}
        auth_mixin.request = Mock(**attrs)
        auth_mixin.arguments["token"] = "NeverHeardOf"
        auth_mixin.prepare()
        assert auth_mixin.is_prepared is False
        assert auth_mixin.status_code == 401

    def test_unset_client_token_with_options(self, auth_mixin):
        """No token is needed for an OPTIONS request. Enables CORS."""
        attrs = {"method": "OPTIONS"}
        auth_mixin.request = Mock(**attrs)
        auth_mixin.prepare()
        assert auth_mixin.is_prepared is True
        assert auth_mixin.status_code is None


class CustomJSONErrorsHandler(JSONErrorsMixin):
    """Implementation that uses the JSONErrorsMixin for testing."""

    def __init__(self):
        self.headers = {}
        self.response = None
        self.status_code = None
        self.reason = None

    def finish(self, response):
        self.response = response

    def set_status(self, status_code, reason=None):
        # The reason parameter is essentially ignored, but necessary in the signature
        # in order for custom messages to be returned to the client from tornado.
        # Tornado's set_status method takes both parameters setting internal members
        # to both values.  For now, we'll set the member but not use it.
        self.status_code = status_code
        self.reason = reason

    def set_header(self, name, value):
        self.headers[name] = value


@pytest.fixture()
def errors_mixin():
    errors_mixin_instance = CustomJSONErrorsHandler()
    return errors_mixin_instance


class TestJSONErrorsMixin:
    """Unit tests the JSON errors mixin."""

    def test_status(self, errors_mixin):
        """Status should be set on the response."""
        errors_mixin.write_error(404)
        response = json.loads(errors_mixin.response)
        assert errors_mixin.status_code == 404
        assert response["reason"] == "Not Found"
        assert response["message"] == ""

    def test_custom_status(self, errors_mixin):
        """Custom reason from exception should be set in the response."""
        exc = web.HTTPError(500, reason="fake-reason")
        errors_mixin.write_error(500, exc_info=[None, exc])

        response = json.loads(errors_mixin.response)
        assert errors_mixin.status_code == 500
        assert response["reason"] == "fake-reason"
        assert response["message"] == ""

    def test_log_message(self, errors_mixin):
        """Custom message from exception should be set in the response."""
        exc = web.HTTPError(410, log_message="fake-message")
        errors_mixin.write_error(410, exc_info=[None, exc])

        response = json.loads(errors_mixin.response)
        assert errors_mixin.status_code == 410
        assert response["reason"] == "Gone"
        assert response["message"] == "fake-message"



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tests for basic gateway app behavior."""
import os
import ssl

import nbformat

from kernel_gateway import __version__
from kernel_gateway.gatewayapp import KernelGatewayApp

RESOURCES = os.path.join(os.path.dirname(__file__), "resources")


class TestGatewayAppConfig:
    """Tests configuration of the gateway app."""

    def test_config_env_vars(self, monkeypatch):
        """Env vars should be honored for traitlets."""
        # Environment vars are always strings
        monkeypatch.setenv("KG_PORT", "1234")
        monkeypatch.setenv("KG_PORT_RETRIES", "4321")
        monkeypatch.setenv("KG_IP", "1.1.1.1")
        monkeypatch.setenv("KG_AUTH_TOKEN", "fake-token")
        monkeypatch.setenv("KG_ALLOW_CREDENTIALS", "true")
        monkeypatch.setenv("KG_ALLOW_HEADERS", "Authorization")
        monkeypatch.setenv("KG_ALLOW_METHODS", "GET")
        monkeypatch.setenv("KG_ALLOW_ORIGIN", "*")
        monkeypatch.setenv("KG_EXPOSE_HEADERS", "X-Fake-Header")
        monkeypatch.setenv("KG_MAX_AGE", "5")
        monkeypatch.setenv("KG_BASE_URL", "/fake/path")
        monkeypatch.setenv("KG_MAX_KERNELS", "1")
        monkeypatch.setenv("KG_SEED_URI", "fake-notebook.ipynb")
        monkeypatch.setenv("KG_PRESPAWN_COUNT", "1")
        monkeypatch.setenv("KG_FORCE_KERNEL_NAME", "fake_kernel_forced")
        monkeypatch.setenv("KG_DEFAULT_KERNEL_NAME", "fake_kernel")
        monkeypatch.setenv("KG_KEYFILE", "/test/fake.key")
        monkeypatch.setenv("KG_CERTFILE", "/test/fake.crt")
        monkeypatch.setenv("KG_CLIENT_CA", "/test/fake_ca.crt")
        monkeypatch.setenv("KG_SSL_VERSION", "3")
        monkeypatch.setenv("KG_TRUST_XHEADERS", "false")

        app = KernelGatewayApp()

        assert app.port == 1234
        assert app.port_retries == 4321
        assert app.ip == "1.1.1.1"
        assert app.auth_token == "fake-token"
        assert app.allow_credentials == "true"
        assert app.allow_headers == "Authorization"
        assert app.allow_methods == "GET"
        assert app.allow_origin == "*"
        assert app.expose_headers == "X-Fake-Header"
        assert app.max_age == "5"
        assert app.base_url == "/fake/path"
        assert app.max_kernels == 1
        assert app.seed_uri == "fake-notebook.ipynb"
        assert app.prespawn_count == 1
        assert app.default_kernel_name == "fake_kernel"
        assert app.force_kernel_name == "fake_kernel_forced"
        assert app.keyfile == "/test/fake.key"
        assert app.certfile == "/test/fake.crt"
        assert app.client_ca == "/test/fake_ca.crt"
        assert app.ssl_version == 3
        assert app.trust_xheaders is False
        KernelGatewayApp.clear_instance()

    def test_trust_xheaders(self, monkeypatch):
        app = KernelGatewayApp()
        assert app.trust_xheaders is False
        monkeypatch.setenv("KG_TRUST_XHEADERS", "true")
        app = KernelGatewayApp()
        assert app.trust_xheaders is True
        KernelGatewayApp.clear_instance()

    def test_ssl_options(self, monkeypatch):
        app = KernelGatewayApp()
        ssl_options = app._build_ssl_options()
        assert ssl_options is None
        KernelGatewayApp.clear_instance()

        # Set all options
        monkeypatch.setenv("KG_CERTFILE", "/test/fake.crt")
        monkeypatch.setenv("KG_KEYFILE", "/test/fake.key")
        monkeypatch.setenv("KG_CLIENT_CA", "/test/fake.ca")
        monkeypatch.setenv("KG_SSL_VERSION", "42")
        app = KernelGatewayApp()
        ssl_options = app._build_ssl_options()
        assert ssl_options["certfile"] == "/test/fake.crt"
        assert ssl_options["keyfile"] == "/test/fake.key"
        assert ssl_options["ca_certs"] == "/test/fake.ca"
        assert ssl_options["cert_reqs"] == ssl.CERT_REQUIRED
        assert ssl_options["ssl_version"] == 42
        KernelGatewayApp.clear_instance()

        # Set few options
        monkeypatch.delenv("KG_KEYFILE")
        monkeypatch.delenv("KG_CLIENT_CA")
        monkeypatch.delenv("KG_SSL_VERSION")
        app = KernelGatewayApp()
        ssl_options = app._build_ssl_options()
        assert ssl_options["certfile"] == "/test/fake.crt"
        assert ssl_options["ssl_version"] == ssl.PROTOCOL_TLSv1_2
        assert "cert_reqs" not in ssl_options
        KernelGatewayApp.clear_instance()

    def test_load_notebook_local(self, monkeypatch):
        nb_path = os.path.join(RESOURCES, "weirdly%20named#notebook.ipynb")
        monkeypatch.setenv("KG_SEED_URI", nb_path)
        with open(nb_path) as nb_fh:
            nb_contents = nbformat.read(nb_fh, 4)

        app = KernelGatewayApp()
        app.init_io_loop()
        app.init_configurables()
        assert app.seed_notebook == nb_contents
        KernelGatewayApp.clear_instance()

    def test_start_banner(self, capsys):
        app = KernelGatewayApp()
        app.init_io_loop()
        app.init_configurables()
        app.start_app()
        log = capsys.readouterr()
        assert f"Jupyter Kernel Gateway {__version__}" in log.err
        KernelGatewayApp.clear_instance()



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tests for notebook-http mode."""

import asyncio
import json
import os

import pytest
from tornado.httpclient import HTTPClientError
from traitlets.config import Config

from kernel_gateway.notebook_http.swagger.handlers import SwaggerSpecHandler

from .test_gatewayapp import RESOURCES


@pytest.fixture()
def jp_server_config():
    """Allows tests to setup their specific configuration values."""
    config = {
        "KernelGatewayApp": {
            "api": "kernel_gateway.notebook_http",
            "seed_uri": os.path.join(RESOURCES, "kernel_api.ipynb"),
        }
    }
    return Config(config)


class TestDefaults:
    """Tests gateway behavior."""

    async def test_api_get_endpoint(self, jp_fetch):
        """GET HTTP method should be callable"""
        response = await jp_fetch("hello", method="GET")
        assert response.code == 200, "GET endpoint did not return 200."
        assert response.body == b"hello world\n", "Unexpected body in response to GET."

    async def test_api_get_endpoint_with_path_param(self, jp_fetch):
        """GET HTTP method should be callable with a path param"""
        response = await jp_fetch("hello", "governor", method="GET")
        assert response.code == 200, "GET endpoint did not return 200."
        assert response.body == b"hello governor\n", "Unexpected body in response to GET."

    async def test_api_get_endpoint_with_query_param(self, jp_fetch):
        """GET HTTP method should be callable with a query param"""
        response = await jp_fetch("hello", "person", params={"person": "governor"}, method="GET")
        assert response.code == 200, "GET endpoint did not return 200."
        print(f"response.body = '{response.body}'")
        assert response.body == b"hello governor\n", "Unexpected body in response to GET."

    async def test_api_get_endpoint_with_multiple_query_params(self, jp_fetch):
        """GET HTTP method should be callable with multiple query params"""
        response = await jp_fetch(
            "hello", "persons", params={"person": "governor, rick"}, method="GET"
        )
        assert response.code == 200, "GET endpoint did not return 200."
        assert response.body == b"hello governor, rick\n", "Unexpected body in response to GET."

    async def test_api_put_endpoint(self, jp_fetch):
        """PUT HTTP method should be callable"""
        response = await jp_fetch("message", method="PUT", body="hola {}")
        assert response.code == 200, "PUT endpoint did not return 200."

        response = await jp_fetch("message", method="GET")
        assert response.code == 200, "GET endpoint did not return 200."
        assert (
            response.body == b"hola {}\n"
        ), "Unexpected body in response to GET after performing PUT."

    async def test_api_post_endpoint(self, jp_fetch):
        """POST endpoint should be callable"""
        expected = b'["Rick", "Maggie", "Glenn", "Carol", "Daryl"]\n'
        response = await jp_fetch(
            "people",
            method="POST",
            body=expected.decode("UTF-8"),
            headers={"Content-Type": "application/json"},
        )
        assert response.code == 200, "POST endpoint did not return 200."
        assert response.body == expected, "Unexpected body in response to POST."

    async def test_api_delete_endpoint(self, jp_fetch):
        """DELETE HTTP method should be callable"""
        expected = b'["Rick", "Maggie", "Glenn", "Carol", "Daryl"]\n'
        response = await jp_fetch(
            "people",
            method="POST",
            body=expected.decode("UTF-8"),
            headers={"Content-Type": "application/json"},
        )
        response = await jp_fetch("people", "2", method="DELETE")
        assert response.code == 200, "DELETE endpoint did not return 200."
        assert (
            response.body == b'["Rick", "Maggie", "Carol", "Daryl"]\n'
        ), "Unexpected body in response to DELETE."

    async def test_api_error_endpoint(self, jp_fetch):
        """Error in a cell should cause 500 HTTP status"""
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("error", method="GET")
        assert e.value.code == 500, "Cell with error did not return 500 status code."

    async def test_api_stderr_endpoint(self, jp_fetch):
        """stderr output in a cell should be dropped"""
        response = await jp_fetch("stderr", method="GET")
        assert response.body == b"I am text on stdout\n", "Unexpected text in response"

    async def test_api_unsupported_method(self, jp_fetch):
        """Endpoints which do no support an HTTP verb should respond with 405."""
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("message", method="DELETE")
        assert (
            e.value.code == 405
        ), "Endpoint which exists, but does not support DELETE, did not return 405 status code."

    async def test_api_undefined(self, jp_fetch):
        """Endpoints which are not registered at all should respond with 404."""
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("not", "an", "endpoint", method="GET")

        assert (
            e.value.code == 404
        ), "Endpoint which should not exist did not return 404 status code."
        body = json.loads(e.value.response.body.decode("UTF-8"))
        assert body["reason"] == "Not Found"

    async def test_api_access_http_header(self, jp_fetch):
        """HTTP endpoints should be able to access request headers"""
        content_types = ["text/plain", "application/json", "application/atom+xml", "foo"]
        for content_type in content_types:
            response = await jp_fetch(
                "content-type", method="GET", headers={"Content-Type": content_type}
            )
            assert response.code == 200, "GET endpoint did not return 200."
            assert (
                response.body.decode(encoding="UTF-8") == f"{content_type}\n"
            ), "Unexpected value in response"

    async def test_format_request_code_escaped_integration(self, jp_fetch):
        """Quotes should be properly escaped in request headers."""
        # Test query with escaping of arguments and headers with multiple escaped quotes
        response = await jp_fetch(
            "hello",
            "person",
            params={"person": "governor"},
            method="GET",
            headers={"If-None-Match": '""9a28a9262f954494a8de7442c63d6d0715ce0998""'},
        )
        assert response.code == 200, "GET endpoint did not return 200."
        assert response.body == b"hello governor\n", "Unexpected body in response to GET."

    async def test_blocked_download_notebook_source(self, jp_fetch):
        """Notebook source should not exist under the path /_api/source when
        `allow_notebook_download` is False or not configured.
        """
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("_api", "source", method="GET")
        assert e.value.code == 404, "/_api/source found when allow_notebook_download is false"

    async def test_blocked_public(self, jp_fetch):
        """Public static assets should not exist under the path /public when
        `static_path` is False or not configured.
        """
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("public", method="GET")
        assert e.value.code == 404, "/public found when static_path is false"

    async def test_api_returns_execute_result(self, jp_fetch):
        """GET HTTP method should return the result of cell execution"""
        response = await jp_fetch("execute_result", method="GET")
        assert response.code == 200, "GET endpoint did not return 200."
        assert response.body == b'{"text/plain": "2"}', "Unexpected body in response to GET."

    async def test_cells_concatenate(self, jp_fetch):
        """Multiple cells with the same verb and path should concatenate."""
        response = await jp_fetch("multi", method="GET")
        assert response.code == 200, "GET endpoint did not return 200."
        assert response.body == b"x is 1\n", "Unexpected body in response to GET."

    async def test_kernel_gateway_environment_set(self, jp_fetch):
        """GET HTTP method should be callable with multiple query params"""
        response = await jp_fetch("env_kernel_gateway", method="GET")
        assert response.code == 200, "GET endpoint did not return 200."
        assert response.body == b"KERNEL_GATEWAY is 1\n", "Unexpected body in response to GET."


@pytest.mark.parametrize(
    "jp_argv", ([f"--NotebookHTTPPersonality.static_path={os.path.join(RESOURCES, 'public')}"],)
)
class TestPublicStatic:
    """Tests gateway behavior when public static assets are enabled."""

    async def test_get_public(self, jp_fetch, jp_argv):
        """index.html should exist under `/public/index.html`."""
        response = await jp_fetch("public", "index.html", method="GET")
        assert response.code == 200
        assert response.headers.get("Content-Type") == "text/html"


@pytest.mark.parametrize("jp_argv", (["--NotebookHTTPPersonality.allow_notebook_download=True"],))
class TestSourceDownload:
    """Tests gateway behavior when notebook download is allowed."""

    async def test_download_notebook_source(self, jp_fetch, jp_argv):
        """Notebook source should exist under the path `/_api/source`."""
        response = await jp_fetch("_api", "source", method="GET")
        assert response.code == 200, "/_api/source did not correctly return the downloaded notebook"
        nb = json.loads(response.body)
        for key in ["cells", "metadata", "nbformat", "nbformat_minor"]:
            assert key in nb


@pytest.mark.parametrize(
    "jp_argv", ([f"--KernelGatewayApp.seed_uri={os.path.join(RESOURCES, 'responses.ipynb')}"],)
)
class TestCustomResponse:
    """Tests gateway behavior when the notebook contains ResponseInfo cells."""

    async def test_setting_content_type(self, jp_fetch, jp_argv):
        """A response cell should allow the content type to be set"""
        response = await jp_fetch("json", method="GET")
        result = json.loads(response.body.decode("UTF-8"))
        assert response.code == 200, "Response status was not 200"
        assert (
            response.headers["Content-Type"] == "application/json"
        ), "Incorrect mime type was set on response"
        assert result == {"hello": "world"}, "Incorrect response value."

    async def test_setting_response_status_code(self, jp_fetch, jp_argv):
        """A response cell should allow the response status code to be set"""
        response = await jp_fetch("nocontent", method="GET")
        assert response.code == 204, "Response status was not 204"
        assert response.body == b"", "Incorrect response value."

    async def test_setting_etag_header(self, jp_fetch, jp_argv):
        """A response cell should allow the etag header to be set"""
        response = await jp_fetch("etag", method="GET")
        result = json.loads(response.body.decode("UTF-8"))
        assert response.code == 200, "Response status was not 200"
        assert (
            response.headers["Content-Type"] == "application/json"
        ), "Incorrect mime type was set on response"
        assert result, {"hello": "world"} == "Incorrect response value."
        assert response.headers["Etag"] == "1234567890", "Incorrect Etag header value."


@pytest.mark.parametrize("jp_argv", (["--KernelGatewayApp.prespawn_count=3"],))
class TestKernelPool:
    async def test_should_cycle_through_kernels(self, jp_fetch, jp_argv):
        """Requests should cycle through kernels"""
        response = await jp_fetch("message", method="PUT", body="hola {}")
        assert response.code == 200, "PUT endpoint did not return 200."

        for i in range(3):
            response = await jp_fetch("message", method="GET")

            if i != 2:
                assert (
                    response.body == b"hello {}\n"
                ), "Unexpected body in response to GET after performing PUT."
            else:
                assert (
                    response.body == b"hola {}\n"
                ), "Unexpected body in response to GET after performing PUT."

    @pytest.mark.timeout(20)
    async def test_concurrent_request_should_not_be_blocked(self, jp_fetch, jp_argv):
        """Concurrent requests should not be blocked"""
        response_long_running = jp_fetch("sleep", "6", method="GET")
        assert response_long_running.done() is False, "Long HTTP Request is not running"

        response_short_running = await jp_fetch("sleep", "3", method="GET")
        assert (
            response_short_running.code == 200
        ), "Short HTTP Request did not return proper status code of 200"
        assert response_long_running.done() is False, "Long HTTP Request is not running"
        while not response_long_running.done():
            await asyncio.sleep(0.3)  # let the long request complete

    async def test_locking_semaphore_of_kernel_resources(self, jp_fetch, jp_argv):
        """Kernel pool should prevent more than one request from running on a kernel at a time."""
        futures = []
        for _ in range(7):
            futures.append(jp_fetch("sleep", "1", method="GET"))

        count = 0
        for future in futures:
            await future
            count += 1
            if count >= 4:
                break

        for future in futures:
            await future


@pytest.mark.parametrize(
    "jp_argv", ([f"--KernelGatewayApp.seed_uri={os.path.join(RESOURCES, 'simple_api.ipynb')}"],)
)
class TestSwaggerSpec:
    async def test_generation_of_swagger_spec(self, jp_fetch, jp_argv):
        """Server should expose a swagger specification of its notebook-defined
        API.
        """
        expected_response = {
            "info": {"version": "0.0.0", "title": "simple_api"},
            "paths": {
                "/name": {
                    "get": {"responses": {"200": {"description": "Success"}}},
                    "post": {"responses": {"200": {"description": "Success"}}},
                }
            },
            "swagger": "2.0",
        }

        response = await jp_fetch("_api", "spec", "swagger.json", method="GET")
        result = json.loads(response.body.decode("UTF-8"))
        assert response.code == 200, "Swagger spec endpoint did not return the correct status code"
        assert result == expected_response, "Swagger spec endpoint did not return the correct value"
        assert (
            SwaggerSpecHandler.output is not None
        ), "Swagger spec output wasn't cached for later requests"


@pytest.mark.parametrize(
    "jp_argv",
    (
        [
            f"--KernelGatewayApp.seed_uri={os.path.join(RESOURCES, 'unknown_kernel.ipynb')}",
            "--KernelGatewayApp.force_kernel_name=python3",
        ],
    ),
)
class TestForceKernel:
    async def test_force_kernel_spec(self, jp_fetch, jp_argv):
        """Should start properly.."""
        response = await jp_fetch("_api", "spec", "swagger.json", method="GET")
        assert response.code == 200



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import logging
import os
from binascii import hexlify

# isort: off
# This must come before any Jupyter imports.
os.environ["JUPYTER_PLATFORM_DIRS"] = "1"
# isort: on

import pytest  # noqa: E402
from traitlets.config import Config  # noqa: E402

from kernel_gateway.gatewayapp import KernelGatewayApp  # noqa: E402

pytest_plugins = ["pytest_jupyter.jupyter_core", "pytest_jupyter.jupyter_server"]


@pytest.fixture(scope="function")
def jp_configurable_serverapp(
    jp_nbconvert_templates,  # this fixture must precede jp_environ
    jp_environ,
    jp_server_config,
    jp_argv,
    jp_http_port,
    jp_base_url,
    tmp_path,
    jp_root_dir,
    jp_logging_stream,
    jp_asyncio_loop,
    io_loop,
):
    """Starts a Jupyter Server instance based on
    the provided configuration values.
    The fixture is a factory; it can be called like
    a function inside a unit test. Here's a basic
    example of how use this fixture:

    .. code-block:: python

      def my_test(jp_configurable_serverapp):
          app = jp_configurable_serverapp(...)
          ...
    """
    KernelGatewayApp.clear_instance()

    def _configurable_serverapp(
        config=jp_server_config,
        base_url=jp_base_url,
        argv=jp_argv,
        http_port=jp_http_port,
        **kwargs,
    ):
        c = Config(config)

        if "auth_token" not in c.KernelGatewayApp and not c.IdentityProvider.token:
            default_token = hexlify(os.urandom(4)).decode("ascii")
            c.IdentityProvider.token = default_token

        app = KernelGatewayApp.instance(
            # Set the log level to debug for testing purposes
            log_level="DEBUG",
            port=http_port,
            port_retries=0,
            base_url=base_url,
            config=c,
            **kwargs,
        )
        app.log.propagate = True
        app.log.handlers = []
        # Initialize app without httpserver
        if jp_asyncio_loop.is_running():
            app.initialize(argv=argv, new_httpserver=False)
        else:

            async def initialize_app():
                app.initialize(argv=argv, new_httpserver=False)

            jp_asyncio_loop.run_until_complete(initialize_app())
        # Reroute all logging StreamHandlers away from stdin/stdout since pytest hijacks
        # these streams and closes them at unfortunate times.
        stream_handlers = [h for h in app.log.handlers if isinstance(h, logging.StreamHandler)]
        for handler in stream_handlers:
            handler.setStream(jp_logging_stream)
        app.log.propagate = True
        app.log.handlers = []
        app.start_app()
        return app

    return _configurable_serverapp


@pytest.fixture(autouse=True)
def jp_server_cleanup(jp_asyncio_loop):
    yield
    app: KernelGatewayApp = KernelGatewayApp.instance()
    try:
        jp_asyncio_loop.run_until_complete(app.async_shutdown())
    except (RuntimeError, SystemExit) as e:
        print("ignoring cleanup error", e)  # noqa: T201
    if hasattr(app, "kernel_manager"):
        app.kernel_manager.context.destroy()
    KernelGatewayApp.clear_instance()


@pytest.fixture()
def jp_auth_header(jp_serverapp):
    """Configures an authorization header using the token from the serverapp fixture."""
    return {"Authorization": f"token {jp_serverapp.identity_provider.token}"}



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tests for jupyter-websocket mode."""

import json
import os
import uuid

import pytest
from jupyter_client.kernelspec import NoSuchKernel
from tornado.escape import json_decode, json_encode, url_escape
from tornado.gen import sleep
from tornado.httpclient import HTTPClientError
from tornado.web import HTTPError
from traitlets.config import Config

from kernel_gateway.gatewayapp import KernelGatewayApp
from kernel_gateway.services.kernels.manager import AsyncMappingKernelManager
from kernel_gateway.services.sessions.sessionmanager import SessionManager

from .test_gatewayapp import RESOURCES


@pytest.fixture()
def jp_server_config():
    """Allows tests to setup their specific configuration values."""
    config = {
        "KernelGatewayApp": {
            "api": "kernel_gateway.jupyter_websocket",
        }
    }
    return Config(config)


@pytest.fixture()
def spawn_kernel(jp_fetch, jp_http_port, jp_base_url, jp_ws_fetch):
    """Spawns a kernel where request.param contains the request body and returns the websocket."""

    async def _spawn_kernel(body="{}"):
        # Request a kernel
        response = await jp_fetch("api", "kernels", method="POST", body=body)
        assert response.code == 201

        # Connect to the kernel via websocket
        kernel = json_decode(response.body)
        kernel_id = kernel["id"]
        ws = await jp_ws_fetch("api", "kernels", kernel_id, "channels")
        return ws

    return _spawn_kernel


def get_execute_request(code: str) -> dict:
    """Creates an execute_request message.

    Parameters
    ----------
    code : str
        Code to execute

    Returns
    -------
    dict
        The message
    """
    return {
        "header": {
            "username": "",
            "version": "5.0",
            "session": "",
            "msg_id": "fake-msg-id",
            "msg_type": "execute_request",
        },
        "parent_header": {},
        "channel": "shell",
        "content": {"code": code, "silent": False, "store_history": False, "user_expressions": {}},
        "metadata": {},
        "buffers": {},
    }


async def await_stream(ws):
    """Returns stream output associated with an execute_request."""
    while 1:
        msg = await ws.read_message()
        msg = json_decode(msg)
        msg_type = msg["msg_type"]
        parent_msg_id = msg["parent_header"]["msg_id"]
        if msg_type == "stream" and parent_msg_id == "fake-msg-id":
            return msg["content"]


class TestDefaults:
    """Tests gateway behavior."""

    @pytest.mark.parametrize("jp_argv", (["--JupyterWebsocketPersonality.list_kernels=True"],))
    async def test_startup(self, jp_fetch, jp_argv):
        """Root of kernels resource should be OK."""
        response = await jp_fetch("api", "kernels", method="GET")
        assert response.code == 200

    async def test_headless(self, jp_fetch):
        """Other notebook resources should not exist."""
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("api", "contents", method="GET")
        assert e.value.code == 404

        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("", method="GET")
        assert e.value.code == 404

        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("tree", method="GET")
        assert e.value.code == 404

    async def test_check_origin(self, jp_fetch, jp_web_app):
        """Allow origin setting should pass through to base handlers."""
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("api", "kernelspecs", headers={"Origin": "fake.com:8888"}, method="GET")
        assert e.value.code == 404

        jp_web_app.settings["allow_origin"] = "*"

        response = await jp_fetch(
            "api", "kernelspecs", headers={"Origin": "fake.com:8888"}, method="GET"
        )
        assert response.code == 200

    @pytest.mark.parametrize(
        "jp_server_config",
        (
            Config(
                {
                    "KernelGatewayApp": {
                        "api": "notebook-gopher",
                    }
                }
            ),
        ),
    )
    async def test_config_bad_api_value(self, jp_configurable_serverapp, jp_server_config):
        """Should raise an ImportError for nonexistent API personality modules."""
        with pytest.raises(ImportError):
            await jp_configurable_serverapp()

    async def test_options_without_auth_token(self, jp_fetch, jp_web_app):
        """OPTIONS requests doesn't need to submit a token. Used for CORS preflight."""
        # Confirm that OPTIONS request doesn't require token
        response = await jp_fetch("api", method="OPTIONS")
        assert response.code == 200

    @pytest.mark.parametrize(
        "jp_server_config",
        (
            Config(
                {
                    "KernelGatewayApp": {
                        "auth_token": "fake-token",
                    }
                }
            ),
        ),
    )
    async def test_auth_token(self, jp_server_config, jp_fetch, jp_web_app, jp_ws_fetch):
        """All server endpoints should check the configured auth token."""

        # Request API without the token
        # Note that we'd prefer not to set _any_ header, but `fp_auth_header` will force it
        # to be set, so setting the empty authorization header is necessary for the tests
        # asserting 401.
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("api", method="GET", headers={"Authorization": ""})
        assert e.value.response.code == 401

        # Now with it
        response = await jp_fetch(
            "api", method="GET", headers={"Authorization": "token fake-token"}
        )
        assert response.code == 200

        # Request kernelspecs without the token
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("api", "kernelspecs", method="GET", headers={"Authorization": ""})
        assert e.value.response.code == 401

        # Now with it
        response = await jp_fetch(
            "api", "kernelspecs", method="GET", headers={"Authorization": "token fake-token"}
        )
        assert response.code == 200

        # Request a kernel without the token
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch(
                "api", "kernels", method="POST", body="{}", headers={"Authorization": ""}
            )
        assert e.value.response.code == 401

        # Now with it
        response = await jp_fetch(
            "api",
            "kernels",
            method="POST",
            body="{}",
            headers={"Authorization": "token fake-token"},
        )
        assert response.code == 201
        kernel = json_decode(response.body)
        kernel_id = url_escape(kernel["id"])

        # Request kernel info without the token
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("api", "kernels", kernel_id, method="GET", headers={"Authorization": ""})
        assert e.value.response.code == 401

        # Now with it
        response = await jp_fetch(
            "api", "kernels", kernel_id, method="GET", headers={"Authorization": "token fake-token"}
        )
        assert response.code == 200

        # Request websocket connection without the token

        # No option to ignore errors so try/except
        with pytest.raises(HTTPClientError) as e:
            await jp_ws_fetch(
                "api", "kernels", kernel_id, "channels", headers={"Authorization": ""}
            )
        assert e.value.response.code == 401

        # Now request the websocket with the token
        ws = await jp_ws_fetch(
            "api", "kernels", kernel_id, "channels", headers={"Authorization": "token fake-token"}
        )
        ws.close()

    async def test_cors_headers(self, jp_fetch, jp_web_app):
        """All kernel endpoints should respond with configured CORS headers."""

        jp_web_app.settings["kg_allow_credentials"] = "false"
        jp_web_app.settings["kg_allow_headers"] = "Authorization,Content-Type"
        jp_web_app.settings["kg_allow_methods"] = "GET,POST"
        jp_web_app.settings["kg_allow_origin"] = "https://jupyter.org"
        jp_web_app.settings["kg_expose_headers"] = "X-My-Fake-Header"
        jp_web_app.settings["kg_max_age"] = "600"
        jp_web_app.settings["kg_list_kernels"] = True

        # Get kernels to check headers
        response = await jp_fetch("api", "kernels", method="GET")
        assert response.code == 200
        assert response.headers["Access-Control-Allow-Credentials"] == "false"
        assert response.headers["Access-Control-Allow-Headers"] == "Authorization,Content-Type"
        assert response.headers["Access-Control-Allow-Methods"] == "GET,POST"
        assert response.headers["Access-Control-Allow-Origin"] == "https://jupyter.org"
        assert response.headers["Access-Control-Expose-Headers"] == "X-My-Fake-Header"
        assert response.headers["Access-Control-Max-Age"] == "600"
        assert response.headers.get("Content-Security-Policy") is None

    async def test_cors_options_headers(self, jp_fetch, jp_web_app):
        """All preflight OPTIONS requests should return configured headers."""
        jp_web_app.settings["kg_allow_headers"] = "X-XSRFToken"
        jp_web_app.settings["kg_allow_methods"] = "GET,POST,OPTIONS"

        response = await jp_fetch("api", "kernelspecs", method="OPTIONS")
        assert response.code == 200
        assert response.headers["Access-Control-Allow-Methods"] == "GET,POST,OPTIONS"
        assert response.headers["Access-Control-Allow-Headers"] == "X-XSRFToken"

    async def test_max_kernels(self, jp_fetch, jp_web_app):
        """Number of kernels should be limited."""
        jp_web_app.settings["kg_max_kernels"] = 1

        # Request a kernel
        response = await jp_fetch("api", "kernels", method="POST", body="{}")
        assert response.code == 201

        # Request another
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("api", "kernels", method="POST", body="{}")
        assert e.value.response.code == 403

        # Shut down the kernel
        kernel = json_decode(response.body)
        response = await jp_fetch("api", "kernels", url_escape(kernel["id"]), method="DELETE")
        assert response.code == 204

        # Try creation again
        response = await jp_fetch("api", "kernels", method="POST", body="{}")
        assert response.code == 201

    async def test_get_api(self, jp_fetch):
        """Server should respond with the API version metadata."""
        response = await jp_fetch("api", method="GET")
        assert response.code == 200
        info = json_decode(response.body)
        assert "version" in info

    async def test_get_kernelspecs(self, jp_fetch):
        """Server should respond with kernel spec metadata."""
        response = await jp_fetch("api", "kernelspecs", method="GET")
        assert response.code == 200
        specs = json_decode(response.body)
        assert "kernelspecs" in specs
        assert "default" in specs

    async def test_get_kernels(self, jp_fetch, jp_web_app):
        """Server should respond with running kernel information."""
        jp_web_app.settings["kg_list_kernels"] = True
        response = await jp_fetch("api", "kernels", method="GET")
        assert response.code == 200
        kernels = json_decode(response.body)
        assert len(kernels) == 0

        # Launch a kernel
        response = await jp_fetch("api", "kernels", method="POST", body="{}")
        assert response.code == 201
        kernel = json_decode(response.body)

        # Check the list again
        response = await jp_fetch("api", "kernels", method="GET")
        assert response.code == 200
        kernels = json_decode(response.body)
        assert len(kernels) == 1
        assert kernels[0]["id"] == kernel["id"]

    async def test_kernel_comm(self, spawn_kernel):
        """Default kernel should launch and accept commands."""
        ws = await spawn_kernel()

        # Send a request for kernel info
        await ws.write_message(
            json_encode(
                {
                    "header": {
                        "username": "",
                        "version": "5.0",
                        "session": "",
                        "msg_id": "fake-msg-id",
                        "msg_type": "kernel_info_request",
                    },
                    "parent_header": {},
                    "channel": "shell",
                    "content": {},
                    "metadata": {},
                    "buffers": {},
                }
            )
        )

        # Assert the reply comes back. Test will timeout if this hangs.
        for _ in range(10):
            msg = await ws.read_message()
            msg = json_decode(msg)
            if msg["msg_type"] == "kernel_info_reply":
                break
        else:
            raise AssertionError("never received kernel_info_reply")
        ws.close()

    async def test_no_discovery(self, jp_fetch):
        """The list of kernels / sessions should be forbidden by default."""
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("api", "kernels", method="GET")
        assert e.value.response.code == 403

        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("api", "sessions", method="GET")
        assert e.value.response.code == 403

    async def test_crud_sessions(self, jp_fetch, jp_web_app):
        """Server should create, list, and delete sessions."""
        jp_web_app.settings["kg_list_kernels"] = True

        # Ensure no sessions by default
        response = await jp_fetch("api", "sessions", method="GET")
        assert response.code == 200
        sessions = json_decode(response.body)
        assert len(sessions) == 0

        # Launch a session
        response = await jp_fetch(
            "api",
            "sessions",
            method="POST",
            body='{"id":"any","notebook":{"path":"anywhere"},"kernel":{"name":"python"}}',
        )
        assert response.code == 201
        session = json_decode(response.body)

        # Check the list again
        response = await jp_fetch("api", "sessions", method="GET")
        assert response.code == 200
        sessions = json_decode(response.body)
        assert len(sessions) == 1
        assert sessions[0]["id"] == session["id"]

        # Delete the session
        response = await jp_fetch("api", "sessions", session["id"], method="DELETE")
        assert response.code == 204

        # Make sure the list is empty
        response = await jp_fetch("api", "sessions", method="GET")
        assert response.code == 200
        sessions = json_decode(response.body)
        assert len(sessions) == 0

    async def test_json_errors(self, jp_fetch):
        """Handlers should always return JSON errors."""
        # A handler that we override
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("api", "kernels", method="GET")
        assert e.value.response.code == 403

        body = json_decode(e.value.response.body)
        assert body["reason"] == "Forbidden"

        # A handler from the notebook base
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("api", "kernels", "1-2-3-4-5", method="GET")
        assert e.value.response.code == 404

        body = json_decode(e.value.response.body)
        assert "1-2-3-4-5" in body["message"]

        # The last resort not found handler
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("fake-endpoint", method="GET")
        assert e.value.response.code == 404

        body = json_decode(e.value.response.body)
        assert body["reason"] == "Not Found"

    @pytest.mark.parametrize("jp_argv", (["--JupyterWebsocketPersonality.env_whitelist=TEST_VAR"],))
    async def test_kernel_env(self, spawn_kernel, jp_argv):
        """Kernel should start with environment vars defined in the request."""

        kernel_body = json.dumps(
            {
                "name": "python",
                "env": {
                    "KERNEL_FOO": "kernel-foo-value",
                    "NOT_KERNEL": "ignored",
                    "KERNEL_GATEWAY": "overridden",
                    "TEST_VAR": "allowed",
                },
            }
        )
        ws = await spawn_kernel(kernel_body)
        req = get_execute_request(
            'import os; print(os.getenv("KERNEL_FOO"), os.getenv("NOT_KERNEL"), '
            'os.getenv("KERNEL_GATEWAY"), os.getenv("TEST_VAR"))'
        )

        await ws.write_message(json_encode(req))
        content = await await_stream(ws)

        assert content["name"] == "stdout"
        assert "kernel-foo-value" in content["text"]
        assert "ignored" not in content["text"]
        assert "overridden" not in content["text"]
        assert "allowed" in content["text"]
        ws.close()

    async def test_get_swagger_yaml_spec(self, jp_fetch):
        """Getting the swagger.yaml spec should be ok"""
        response = await jp_fetch("api", "swagger.yaml", method="GET")
        assert response.code == 200

    async def test_get_swagger_json_spec(self, jp_fetch):
        """Getting the swagger.json spec should be ok"""
        response = await jp_fetch("api", "swagger.json", method="GET")
        assert response.code == 200

    async def test_kernel_env_auth_token(self, monkeypatch, spawn_kernel):
        """Kernel should not have KG_AUTH_TOKEN in its environment."""
        monkeypatch.setenv("KG_AUTH_TOKEN", "fake-secret")

        ws = None
        try:
            ws = await spawn_kernel()
            req = get_execute_request("import os; print(os.getenv('KG_AUTH_TOKEN'))")
            await ws.write_message(json_encode(req))
            content = await await_stream(ws)
            assert "fake-secret" not in content["text"]
            assert "None" in content["text"]  # ensure None was printed
        finally:
            if ws is not None:
                ws.close()


@pytest.mark.parametrize("jp_argv", (["--KernelGatewayApp.default_kernel_name=fake-kernel"],))
class TestCustomDefaultKernel:
    """Tests gateway behavior when setting a custom default kernelspec."""

    async def test_default_kernel_name(self, jp_argv, jp_fetch):
        """The default kernel name should be used on empty requests."""
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("api", "kernels", method="POST", body="")
        assert e.value.response.code == 500
        assert "raise NoSuchKernel" in str(e.value.response.body)


@pytest.mark.parametrize(
    "jp_argv",
    (
        [
            "--KernelGatewayApp.prespawn_count=2",
            f"--KernelGatewayApp.seed_uri={os.path.join(RESOURCES, 'zen.ipynb')}",
            "--KernelGatewayApp.force_kernel_name=python3",
        ],
    ),
)
class TestForceKernel:
    """Tests gateway behavior when forcing a kernelspec."""

    async def test_force_kernel_name(self, jp_argv, jp_fetch):
        """Should create a Python kernel."""
        response = await jp_fetch("api", "kernels", method="POST", body='{"name": "fake-kernel"}')
        assert response.code == 201
        kernel = json_decode(response.body)
        assert kernel["name"] == "python3"


class TestEnableDiscovery:
    """Tests gateway behavior with kernel listing enabled."""

    @pytest.mark.parametrize("jp_argv", (["--JupyterWebsocketPersonality.list_kernels=True"],))
    async def test_enable_kernel_list(self, jp_fetch, jp_argv):
        """The list of kernels, sessions, and activities should be available."""

        response = await jp_fetch("api", "kernels", method="GET")
        assert response.code == 200
        assert "[]" in str(response.body)

        response = await jp_fetch("api", "sessions", method="GET")
        assert response.code == 200
        assert "[]" in str(response.body)


class TestPrespawnKernels:
    """Tests gateway behavior when kernels are spawned at startup."""

    @pytest.mark.parametrize("jp_argv", (["--KernelGatewayApp.prespawn_count=2"],))
    async def test_prespawn_count(self, jp_fetch, jp_web_app, jp_argv):
        """Server should launch the given number of kernels on startup."""
        jp_web_app.settings["kg_list_kernels"] = True
        await sleep(0.5)
        response = await jp_fetch("api", "kernels", method="GET")
        assert response.code == 200

        kernels = json_decode(response.body)
        assert len(kernels) == 2

    def test_prespawn_max_conflict(self):
        """Server should error if prespawn count is greater than max allowed kernels."""
        app = KernelGatewayApp()
        app.prespawn_count = 3
        app.max_kernels = 2
        with pytest.raises(RuntimeError):
            app.init_configurables()


class TestBaseURL:
    """Tests gateway behavior when a custom base URL is configured."""

    @pytest.mark.parametrize("jp_argv", (["--JupyterWebsocketPersonality.list_kernels=True"],))
    @pytest.mark.parametrize("jp_base_url", ("/fake/path",))
    async def test_base_url(self, jp_base_url, jp_argv, jp_fetch):
        """Server should mount resources under configured base."""
        # Should exist under path
        response = await jp_fetch("api", "kernels", method="GET")
        assert response.code == 200
        assert "/fake/path/api/kernels" in response.effective_url


class TestRelativeBaseURL:
    """Tests gateway behavior when a relative base URL is configured."""

    @pytest.mark.parametrize("jp_argv", (["--JupyterWebsocketPersonality.list_kernels=True"],))
    @pytest.mark.parametrize("jp_base_url", ("/fake/path",))
    async def test_base_url(self, jp_base_url, jp_argv, jp_fetch):
        """Server should mount resources under fixed base."""

        # Should exist under path
        response = await jp_fetch("api", "kernels", method="GET")
        assert response.code == 200
        assert "/fake/path/api/kernels" in response.effective_url


class TestSeedURI:
    """Tests gateway behavior when a seeding kernel memory with code from a notebook."""

    @pytest.mark.parametrize(
        "jp_argv", ([f"--KernelGatewayApp.seed_uri={os.path.join(RESOURCES, 'zen.ipynb')}"],)
    )
    async def test_seed(self, jp_argv, spawn_kernel):
        """Kernel should have variables pre-seeded from the notebook."""
        ws = await spawn_kernel()

        # Print the encoded "zen of python" string, the kernel should have
        # it imported
        req = get_execute_request("print(this.s)")
        await ws.write_message(json_encode(req))
        content = await await_stream(ws)
        assert content["name"] == "stdout"
        assert "Gur Mra bs Clguba" in content["text"]

        ws.close()


class TestRemoteSeedURI:
    """Tests gateway behavior when a seeding kernel memory with code from a remote notebook."""

    @pytest.mark.parametrize(
        "jp_argv",
        (
            [
                "--KernelGatewayApp.seed_uri="
                "https://gist.githubusercontent.com/parente/ccd36bd7db2f617d58ce/raw/zen3.ipynb"
            ],
        ),
    )
    async def test_seed(self, jp_argv, spawn_kernel):
        """Kernel should have variables pre-seeded from the notebook."""
        ws = await spawn_kernel()

        # Print the encoded "zen of python" string, the kernel should have
        # it imported
        req = get_execute_request("print(this.s)")
        await ws.write_message(json_encode(req))
        content = await await_stream(ws)
        assert content["name"] == "stdout"
        assert "Gur Mra bs Clguba" in content["text"]

        ws.close()


class TestBadSeedURI:
    """Tests gateway behavior when seeding kernel memory with notebook code that fails."""

    @pytest.mark.parametrize(
        "jp_argv",
        (
            [
                f"--KernelGatewayApp.seed_uri={os.path.join(RESOURCES, 'failing_code.ipynb')}",
                "--JupyterWebsocketPersonality.list_kernels=True",
            ],
        ),
    )
    async def test_seed_error(self, jp_argv, jp_fetch):
        """
        Server should shutdown kernel and respond with error when seed notebook
        has an execution error.
        """

        # Request a kernel
        with pytest.raises(HTTPClientError) as e:
            await jp_fetch("api", "kernels", method="POST", body="{}")
        assert e.value.response.code == 500

        # No kernels should be running
        response = await jp_fetch("api", "kernels", method="GET")
        assert response.code == 200
        kernels = json_decode(response.body)
        assert len(kernels) == 0

    async def test_seed_kernel_not_available(self):
        """
        Server should error because seed notebook requires a kernel that is not installed.
        """
        app = KernelGatewayApp()
        app.seed_uri = os.path.join(RESOURCES, "unknown_kernel.ipynb")
        with pytest.raises(NoSuchKernel):
            app.init_configurables()


@pytest.mark.parametrize(
    "jp_argv",
    (
        [
            f"--KernelGatewayApp.seed_uri={os.path.join(RESOURCES, 'zen.ipynb')}",
            "--KernelGatewayApp.prespawn_count=1",
        ],
    ),
)
class TestKernelLanguageSupport:
    """Tests gateway behavior when a client requests a specific kernel spec."""

    async def test_seed_language_support(self, jp_argv, spawn_kernel):
        """Kernel should have variables pre-seeded from notebook."""
        ws = await spawn_kernel(body=json.dumps({"name": "python3"}))
        code = "print(this.s)"

        # Print the encoded "zen of python" string, the kernel should have it imported
        req = get_execute_request(code)
        await ws.write_message(json_encode(req))
        content = await await_stream(ws)
        assert content["name"] == "stdout"
        assert "Gur Mra bs Clguba" in content["text"]

        ws.close()


class TestSessionApi:
    """Test session object API to improve coverage."""

    async def test_session_api(self, tmp_path, jp_environ):
        # Create the manager instances
        akm = AsyncMappingKernelManager()
        sm = SessionManager(akm)

        row_model = await sm.create_session(path=str(tmp_path), kernel_name="python3")
        assert "id" in row_model
        assert "kernel" in row_model
        assert row_model["notebook"]["path"] == str(tmp_path)

        session_id = row_model["id"]
        kernel_id = row_model["kernel"]["id"]

        # Perform some get_session tests
        with pytest.raises(TypeError):
            sm.get_session()  # no kwargs

        with pytest.raises(TypeError):
            kwargs = {"bogus_column": 1}
            sm.get_session(**kwargs)  # bad column

        non_existent_session_id = uuid.uuid4()
        with pytest.raises(HTTPError) as e:
            kwargs = {"session_id": str(non_existent_session_id)}
            sm.get_session(**kwargs)  # bad session id
        assert e.value.status_code == 404

        # Perform some update_session tests
        sm.update_session(session_id)  # no kwargs - success expected

        with pytest.raises(KeyError):
            kwargs = {"kernel_id": kernel_id}
            sm.update_session(str(non_existent_session_id), **kwargs)  # bad session id

        kwargs = {"path": "/tmp"}
        sm.update_session(session_id, **kwargs)  # update path of session

        # confirm update
        kwargs = {"session_id": session_id}
        row_model = sm.get_session(**kwargs)
        assert row_model["notebook"]["path"] == "/tmp"

        kwargs = {"kernel_id": str(uuid.uuid4())}
        with pytest.raises(KeyError):
            sm.update_session(session_id, **kwargs)  # bad kernel_id

        await sm.delete_session(session_id)

        with pytest.raises(HTTPError) as e:
            kwargs = {"session_id": session_id}
            sm.get_session(**kwargs)
        assert e.value.status_code == 404



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tornado handlers for kernel specs."""

import os

from tornado import web

from ..mixins import CORSMixin


class BaseSpecHandler(CORSMixin, web.StaticFileHandler):
    """Exposes the ability to return specifications from static files"""

    @staticmethod
    def get_resource_metadata():
        """Returns the (resource, mime-type) for the handlers spec."""

    def initialize(self):
        """Initializes the instance of this class to serve files.

        The handler is initialized to server files from the directory
        where this module is defined.
        """
        web.StaticFileHandler.initialize(self, path=os.path.dirname(__file__))

    async def get(self):
        """Handler for a get on a specific handler"""
        resource_name, content_type = self.get_resource_metadata()
        self.set_header("Content-Type", content_type)
        res = await web.StaticFileHandler.get(self, resource_name)
        return res

    def options(self, **kwargs):
        """Method for properly handling CORS pre-flight"""
        self.finish()


class SpecJsonHandler(BaseSpecHandler):
    """Exposes a JSON swagger specification"""

    @staticmethod
    def get_resource_metadata():
        return "swagger.json", "application/json"


class APIYamlHandler(BaseSpecHandler):
    """Exposes a YAML swagger specification"""

    @staticmethod
    def get_resource_metadata():
        return "swagger.yaml", "text/x-yaml"


default_handlers = [
    (f"/api/{SpecJsonHandler.get_resource_metadata()[0]}", SpecJsonHandler),
    (f"/api/{APIYamlHandler.get_resource_metadata()[0]}", APIYamlHandler),
]



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""CLI entrypoint for the kernel gateway package."""

if __name__ == "__main__":
    import kernel_gateway.gatewayapp as app

    app.launch_instance()



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Kernel Gateway Jupyter application."""

import asyncio
import errno
import hashlib
import hmac
import importlib
import logging
import os
import select
import signal
import ssl
import sys
import threading
from base64 import encodebytes
from urllib.parse import urlparse

import nbformat
from jupyter_client.kernelspec import KernelSpecManager
from jupyter_core.application import JupyterApp, base_aliases
from jupyter_core.paths import secure_write
from jupyter_server.auth.authorizer import AllowAllAuthorizer, Authorizer
from jupyter_server.serverapp import random_ports
from jupyter_server.services.kernels.connection.base import BaseKernelWebsocketConnection
from jupyter_server.services.kernels.connection.channels import ZMQChannelsWebsocketConnection
from jupyter_server.services.kernels.kernelmanager import MappingKernelManager
from tornado import httpserver, ioloop, web
from tornado.log import LogFormatter, enable_pretty_logging
from traitlets import Bytes, CBool, Instance, Integer, List, Type, Unicode, default, observe

from ._version import __version__
from .auth.identity import GatewayIdentityProvider
from .jupyter_websocket import JupyterWebsocketPersonality

# Only present for generating help documentation
from .notebook_http import NotebookHTTPPersonality
from .services.kernels.manager import SeedingMappingKernelManager
from .services.sessions.sessionmanager import SessionManager

# Add additional command line aliases
aliases = dict(base_aliases)
aliases.update(
    {
        "ip": "KernelGatewayApp.ip",
        "port": "KernelGatewayApp.port",
        "port_retries": "KernelGatewayApp.port_retries",
        "api": "KernelGatewayApp.api",
        "seed_uri": "KernelGatewayApp.seed_uri",
        "keyfile": "KernelGatewayApp.keyfile",
        "certfile": "KernelGatewayApp.certfile",
        "client-ca": "KernelGatewayApp.client_ca",
        "ssl_version": "KernelGatewayApp.ssl_version",
    }
)


class KernelGatewayApp(JupyterApp):
    """Application that provisions Jupyter kernels and proxies HTTP/Websocket
    traffic to the kernels.

    - reads command line and environment variable settings
    - initializes managers and routes
    - creates a Tornado HTTP server
    - starts the Tornado event loop
    """

    name = "jupyter-kernel-gateway"
    version = __version__
    description = """
        Jupyter Kernel Gateway

        Provisions Jupyter kernels and proxies HTTP/Websocket traffic
        to them.
    """

    # Also include when generating help options
    classes = [NotebookHTTPPersonality, JupyterWebsocketPersonality]
    # Enable some command line shortcuts
    aliases = aliases

    # Server IP / PORT binding
    port_env = "KG_PORT"
    port_default_value = 8888
    port = Integer(
        port_default_value, config=True, help="Port on which to listen (KG_PORT env var)"
    )

    @default("port")
    def port_default(self):
        return int(os.getenv(self.port_env, self.port_default_value))

    port_retries_env = "KG_PORT_RETRIES"
    port_retries_default_value = 50
    port_retries = Integer(
        port_retries_default_value,
        config=True,
        help="Number of ports to try if the specified port is not available (KG_PORT_RETRIES env var)",
    )

    @default("port_retries")
    def port_retries_default(self):
        return int(os.getenv(self.port_retries_env, self.port_retries_default_value))

    ip_env = "KG_IP"
    ip_default_value = "127.0.0.1"
    ip = Unicode(
        ip_default_value, config=True, help="IP address on which to listen (KG_IP env var)"
    )

    @default("ip")
    def ip_default(self):
        return os.getenv(self.ip_env, self.ip_default_value)

    # Base URL
    base_url_env = "KG_BASE_URL"
    base_url_default_value = "/"
    base_url = Unicode(
        base_url_default_value,
        config=True,
        help="""The base path for mounting all API resources (KG_BASE_URL env var)""",
    )

    @default("base_url")
    def base_url_default(self):
        return os.getenv(self.base_url_env, self.base_url_default_value)

    # Token authorization
    auth_token_env = "KG_AUTH_TOKEN"  # noqa: S105
    auth_token = Unicode(
        config=True, help="Authorization token required for all requests (KG_AUTH_TOKEN env var)"
    )

    @default("auth_token")
    def _auth_token_default(self):
        return os.getenv(self.auth_token_env, "")

    # CORS headers
    allow_credentials_env = "KG_ALLOW_CREDENTIALS"
    allow_credentials = Unicode(
        config=True,
        help="Sets the Access-Control-Allow-Credentials header. (KG_ALLOW_CREDENTIALS env var)",
    )

    @default("allow_credentials")
    def allow_credentials_default(self):
        return os.getenv(self.allow_credentials_env, "")

    allow_headers_env = "KG_ALLOW_HEADERS"
    allow_headers = Unicode(
        config=True, help="Sets the Access-Control-Allow-Headers header. (KG_ALLOW_HEADERS env var)"
    )

    @default("allow_headers")
    def allow_headers_default(self):
        return os.getenv(self.allow_headers_env, "")

    allow_methods_env = "KG_ALLOW_METHODS"
    allow_methods = Unicode(
        config=True, help="Sets the Access-Control-Allow-Methods header. (KG_ALLOW_METHODS env var)"
    )

    @default("allow_methods")
    def allow_methods_default(self):
        return os.getenv(self.allow_methods_env, "")

    allow_origin_env = "KG_ALLOW_ORIGIN"
    allow_origin = Unicode(
        config=True, help="Sets the Access-Control-Allow-Origin header. (KG_ALLOW_ORIGIN env var)"
    )

    @default("allow_origin")
    def allow_origin_default(self):
        return os.getenv(self.allow_origin_env, "")

    expose_headers_env = "KG_EXPOSE_HEADERS"
    expose_headers = Unicode(
        config=True,
        help="Sets the Access-Control-Expose-Headers header. (KG_EXPOSE_HEADERS env var)",
    )

    @default("expose_headers")
    def expose_headers_default(self):
        return os.getenv(self.expose_headers_env, "")

    trust_xheaders_env = "KG_TRUST_XHEADERS"
    trust_xheaders = CBool(
        False,
        config=True,
        help="Use x-* header values for overriding the remote-ip, useful when application is behind a proxy. (KG_TRUST_XHEADERS env var)",
    )

    @default("trust_xheaders")
    def trust_xheaders_default(self):
        return os.getenv(self.trust_xheaders_env, "False").lower() == "true"

    max_age_env = "KG_MAX_AGE"
    max_age = Unicode(
        config=True, help="Sets the Access-Control-Max-Age header. (KG_MAX_AGE env var)"
    )

    @default("max_age")
    def max_age_default(self):
        return os.getenv(self.max_age_env, "")

    max_kernels_env = "KG_MAX_KERNELS"
    max_kernels = Integer(
        None,
        config=True,
        allow_none=True,
        help="Limits the number of kernel instances allowed to run by this gateway. Unbounded by default. (KG_MAX_KERNELS env var)",
    )

    @default("max_kernels")
    def max_kernels_default(self):
        val = os.getenv(self.max_kernels_env)
        return val if val is None else int(val)

    seed_uri_env = "KG_SEED_URI"
    seed_uri = Unicode(
        None,
        config=True,
        allow_none=True,
        help="Runs the notebook (.ipynb) at the given URI on every kernel launched. No seed by default. (KG_SEED_URI env var)",
    )

    @default("seed_uri")
    def seed_uri_default(self):
        return os.getenv(self.seed_uri_env)

    prespawn_count_env = "KG_PRESPAWN_COUNT"
    prespawn_count = Integer(
        None,
        config=True,
        allow_none=True,
        help="Number of kernels to prespawn using the default language. No prespawn by default. (KG_PRESPAWN_COUNT env var)",
    )

    @default("prespawn_count")
    def prespawn_count_default(self):
        val = os.getenv(self.prespawn_count_env)
        return val if val is None else int(val)

    default_kernel_name_env = "KG_DEFAULT_KERNEL_NAME"
    default_kernel_name = Unicode(
        config=True,
        help="Default kernel name when spawning a kernel (KG_DEFAULT_KERNEL_NAME env var)",
    )

    @default("default_kernel_name")
    def default_kernel_name_default(self):
        # defaults to Jupyter's default kernel name on empty string
        return os.getenv(self.default_kernel_name_env, "")

    force_kernel_name_env = "KG_FORCE_KERNEL_NAME"
    force_kernel_name = Unicode(
        config=True,
        help="Override any kernel name specified in a notebook or request (KG_FORCE_KERNEL_NAME env var)",
    )

    @default("force_kernel_name")
    def force_kernel_name_default(self):
        return os.getenv(self.force_kernel_name_env, "")

    env_process_whitelist_env = "KG_ENV_PROCESS_WHITELIST"
    env_process_whitelist = List(
        config=True,
        help="""Environment variables allowed to be inherited from the spawning process by the kernel""",
    )

    @default("env_process_whitelist")
    def env_process_whitelist_default(self):
        return os.getenv(self.env_process_whitelist_env, "").split(",")

    api_env = "KG_API"
    api_default_value = "kernel_gateway.jupyter_websocket"
    api = Unicode(
        api_default_value,
        config=True,
        help="""Controls which API to expose, that of a Jupyter notebook server, the seed
            notebook's, or one provided by another module, respectively using values
            'kernel_gateway.jupyter_websocket', 'kernel_gateway.notebook_http', or
            another fully qualified module name (KG_API env var)
            """,
    )

    @default("api")
    def api_default(self):
        return os.getenv(self.api_env, self.api_default_value)

    @observe("api")
    def api_changed(self, event):
        try:
            self._load_api_module(event["new"])
        except ImportError:
            # re-raise with more sensible message to help the user
            raise ImportError("API module {} not found".format(event["new"])) from None

    certfile_env = "KG_CERTFILE"
    certfile = Unicode(
        None,
        config=True,
        allow_none=True,
        help="""The full path to an SSL/TLS certificate file. (KG_CERTFILE env var)""",
    )

    @default("certfile")
    def certfile_default(self):
        return os.getenv(self.certfile_env)

    keyfile_env = "KG_KEYFILE"
    keyfile = Unicode(
        None,
        config=True,
        allow_none=True,
        help="""The full path to a private key file for usage with SSL/TLS. (KG_KEYFILE env var)""",
    )

    @default("keyfile")
    def keyfile_default(self):
        return os.getenv(self.keyfile_env)

    client_ca_env = "KG_CLIENT_CA"
    client_ca = Unicode(
        None,
        config=True,
        allow_none=True,
        help="""The full path to a certificate authority certificate for SSL/TLS client authentication. (KG_CLIENT_CA env var)""",
    )

    @default("client_ca")
    def client_ca_default(self):
        return os.getenv(self.client_ca_env)

    ssl_version_env = "KG_SSL_VERSION"
    ssl_version_default_value = ssl.PROTOCOL_TLSv1_2
    ssl_version = Integer(
        None,
        config=True,
        allow_none=True,
        help="""Sets the SSL version to use for the web socket connection. (KG_SSL_VERSION env var)""",
    )

    @default("ssl_version")
    def ssl_version_default(self):
        ssl_from_env = os.getenv(self.ssl_version_env)
        return ssl_from_env if ssl_from_env is None else int(ssl_from_env)

    cookie_secret_file = Unicode(
        config=True, help="""The file where the cookie secret is stored."""
    )

    @default("cookie_secret_file")
    def _default_cookie_secret_file(self):
        return os.path.join(self.runtime_dir, "jupyter_cookie_secret")

    cookie_secret = Bytes(
        b"",
        config=True,
        help="""The random bytes used to secure cookies.
        By default this is a new random number every time you start the server.
        Set it to a value in a config file to enable logins to persist across server sessions.

        Note: Cookie secrets should be kept private, do not share config files with
        cookie_secret stored in plaintext (you can read the value from a file).
        """,
    )

    @default("cookie_secret")
    def _default_cookie_secret(self):
        if os.path.exists(self.cookie_secret_file):
            with open(self.cookie_secret_file, "rb") as f:
                key = f.read()
        else:
            key = encodebytes(os.urandom(32))
            self._write_cookie_secret_file(key)
        h = hmac.new(key, digestmod=hashlib.sha256)
        # h.update(self.password.encode())  # password is deprecated in 2.0
        return h.digest()

    def _write_cookie_secret_file(self, secret):
        """write my secret to my secret_file"""
        self.log.info("Writing Jupyter server cookie secret to %s", self.cookie_secret_file)
        try:
            with secure_write(self.cookie_secret_file, True) as f:
                f.write(secret)
        except OSError as e:
            self.log.error(
                "Failed to write cookie secret to %s: %s",
                self.cookie_secret_file,
                e,
            )

    ws_ping_interval_env = "KG_WS_PING_INTERVAL_SECS"
    ws_ping_interval_default_value = 30
    ws_ping_interval = Integer(
        ws_ping_interval_default_value,
        config=True,
        help="""Specifies the ping interval(in seconds) that should be used by zmq port
                                     associated with spawned kernels. Set this variable to 0 to disable ping mechanism.
                                    (KG_WS_PING_INTERVAL_SECS env var)""",
    )

    @default("ws_ping_interval")
    def _ws_ping_interval_default(self) -> int:
        return int(os.getenv(self.ws_ping_interval_env, self.ws_ping_interval_default_value))

    _log_formatter_cls = LogFormatter  # traitlet default is LevelFormatter

    @default("log_format")
    def _default_log_format(self) -> str:
        """override default log format to include milliseconds"""
        return (
            "%(color)s[%(levelname)1.1s %(asctime)s.%(msecs).03d %(name)s]%(end_color)s %(message)s"
        )

    kernel_spec_manager = Instance(KernelSpecManager, allow_none=True)

    kernel_spec_manager_class = Type(
        default_value=KernelSpecManager,
        config=True,
        help="""
        The kernel spec manager class to use. Should be a subclass
        of `jupyter_client.kernelspec.KernelSpecManager`.
        """,
    )

    kernel_manager_class = Type(
        klass=MappingKernelManager,
        default_value=SeedingMappingKernelManager,
        config=True,
        help="""The kernel manager class to use.""",
    )

    kernel_websocket_connection_class = Type(
        default_value=ZMQChannelsWebsocketConnection,
        klass=BaseKernelWebsocketConnection,
        config=True,
        help="""The kernel websocket connection class to use.""",
    )

    authorizer_class = Type(
        default_value=AllowAllAuthorizer,
        klass=Authorizer,
        config=True,
        help="The authorizer class to use.",
    )

    identity_provider_class = Type(
        default_value=GatewayIdentityProvider,
        klass=GatewayIdentityProvider,
        config=True,
        help="The identity provider class to use.",
    )

    def _load_api_module(self, module_name):
        """Tries to import the given module name.

        Parameters
        ----------
        module_name: str
            Module name to import

        Returns
        -------
        module
            Module with the given name loaded using importlib.import_module
        """
        # some compatibility allowances
        if module_name == "jupyter-websocket":
            module_name = "kernel_gateway.jupyter_websocket"
        elif module_name == "notebook-http":
            module_name = "kernel_gateway.notebook_http"
        return importlib.import_module(module_name)

    def _load_notebook(self, uri):
        """Loads a notebook from the local filesystem or HTTP(S) URL.

        Raises
        ------
        RuntimeError if there is no kernel spec matching the one specified in
        the notebook or forced via configuration.

        Returns
        -------
        object
            Notebook object from nbformat
        """
        parts = urlparse(uri)

        if parts.scheme not in ("http", "https"):
            # Local file
            path = parts._replace(scheme="", netloc="").geturl()
            with open(path) as nb_fh:
                notebook = nbformat.read(nb_fh, 4)
        else:
            # Remote file
            import requests

            resp = requests.get(uri, timeout=200)
            resp.raise_for_status()
            notebook = nbformat.reads(resp.text, 4)

        # Error if no kernel spec can handle the language requested
        kernel_name = (
            self.force_kernel_name
            if self.force_kernel_name
            else notebook["metadata"]["kernelspec"]["name"]
        )
        self.kernel_spec_manager.get_kernel_spec(kernel_name)

        return notebook

    def initialize(
        self,
        argv=None,
        new_httpserver=True,
    ):
        """Initializes the base class, configurable manager instances, the
        Tornado web app, and the tornado HTTP server.

        Parameters
        ----------
        argv
            Command line arguments

        new_httpserver
            Indicates that a new HTTP server instance should be created
        """
        super().initialize(argv)

        self.init_io_loop()
        self.init_configurables()
        self.init_webapp()
        self.init_signal()
        if new_httpserver:
            self.init_http_server()

    def init_configurables(self):
        """Initializes all configurable objects including a kernel manager, kernel
        spec manager, session manager, and personality.

        Any kernel pool configured by the personality will be its responsibility
        to shut down.

        Optionally, loads a notebook and prespawns the configured number of
        kernels.
        """
        self.kernel_spec_manager = KernelSpecManager(parent=self)

        self.seed_notebook = None
        if self.seed_uri is not None:
            # Note: must be set before instantiating a SeedingMappingKernelManager
            self.seed_notebook = self._load_notebook(self.seed_uri)

        # Only pass a default kernel name when one is provided. Otherwise,
        # adopt whatever default the kernel manager wants to use.
        kwargs = {}
        if self.default_kernel_name:
            kwargs["default_kernel_name"] = self.default_kernel_name

        self.kernel_spec_manager = self.kernel_spec_manager_class(
            parent=self,
        )
        self.kernel_manager = self.kernel_manager_class(
            parent=self,
            log=self.log,
            connection_dir=self.runtime_dir,
            kernel_spec_manager=self.kernel_spec_manager,
            **kwargs,
        )

        self.session_manager = SessionManager(log=self.log, kernel_manager=self.kernel_manager)
        self.contents_manager = None

        self.identity_provider = self.identity_provider_class(parent=self, log=self.log)

        self.authorizer = self.authorizer_class(
            parent=self, log=self.log, identity_provider=self.identity_provider
        )

        if self.prespawn_count:
            if self.max_kernels and self.prespawn_count > self.max_kernels:
                msg = f"Cannot prespawn {self.prespawn_count} kernels; more than max kernels {self.max_kernels}"
                raise RuntimeError(msg)

        api_module = self._load_api_module(self.api)
        func = api_module.create_personality
        self.personality = func(parent=self, log=self.log)

        self.io_loop.call_later(
            0.1, lambda: asyncio.create_task(self.personality.init_configurables())
        )

    def init_webapp(self):
        """Initializes Tornado web application with uri handlers.

        Adds the various managers and web-front configuration values to the
        Tornado settings for reference by the handlers.
        """
        # Enable the same pretty logging the notebook uses
        enable_pretty_logging()

        # Configure the tornado logging level too
        logging.getLogger().setLevel(self.log_level)

        handlers = self.personality.create_request_handlers()

        self.web_app = web.Application(
            handlers=handlers,
            kernel_manager=self.kernel_manager,
            session_manager=self.session_manager,
            contents_manager=self.contents_manager,
            kernel_spec_manager=self.kernel_spec_manager,
            kg_auth_token=self.auth_token,
            kg_allow_credentials=self.allow_credentials,
            kg_allow_headers=self.allow_headers,
            kg_allow_methods=self.allow_methods,
            kg_allow_origin=self.allow_origin,
            kg_expose_headers=self.expose_headers,
            kg_max_age=self.max_age,
            kg_max_kernels=self.max_kernels,
            kg_env_process_whitelist=self.env_process_whitelist,
            kg_api=self.api,
            kg_personality=self.personality,
            # Also set the allow_origin setting used by notebook so that the
            # check_origin method used everywhere respects the value
            allow_origin=self.allow_origin,
            # Set base_url for use in request handlers
            base_url=self.base_url,
            # Authentication
            cookie_secret=self.cookie_secret,
            # Always allow remote access (has been limited to localhost >= notebook 5.6)
            allow_remote_access=True,
            # setting ws_ping_interval value that can allow it to be modified for the purpose of toggling ping mechanism
            # for zmq web-sockets or increasing/decreasing web socket ping interval/timeouts.
            ws_ping_interval=self.ws_ping_interval * 1000,
            # Add a pass-through authorizer for now
            authorizer=self.authorizer_class(parent=self),
            identity_provider=self.identity_provider,
            kernel_websocket_connection_class=self.kernel_websocket_connection_class,
        )

        # promote the current personality's "config" tagged traitlet values to webapp settings
        for trait_name, trait_value in self.personality.class_traits(config=True).items():
            kg_name = "kg_" + trait_name
            # a personality's traitlets may not overwrite the kernel gateway's
            if kg_name not in self.web_app.settings:
                self.web_app.settings[kg_name] = trait_value.get(obj=self.personality)
            else:
                self.log.warning(
                    "The personality trait name, %s, conflicts with a kernel gateway trait.",
                    trait_name,
                )

    def _build_ssl_options(self):
        """Build a dictionary of SSL options for the tornado HTTP server.

        Taken directly from jupyter/notebook code.
        """
        ssl_options = {}
        if self.certfile:
            ssl_options["certfile"] = self.certfile
        if self.keyfile:
            ssl_options["keyfile"] = self.keyfile
        if self.client_ca:
            ssl_options["ca_certs"] = self.client_ca
        if self.ssl_version:
            ssl_options["ssl_version"] = self.ssl_version
        if not ssl_options:
            # None indicates no SSL config
            ssl_options = None
        else:
            ssl_options.setdefault("ssl_version", self.ssl_version_default_value)
            if ssl_options.get("ca_certs", False):
                ssl_options.setdefault("cert_reqs", ssl.CERT_REQUIRED)

        return ssl_options

    def init_http_server(self):
        """Initializes a HTTP server for the Tornado web application on the
        configured interface and port.

        Tries to find an open port if the one configured is not available using
        the same logic as the Jupyer Notebook server.
        """
        ssl_options = self._build_ssl_options()
        self.http_server = httpserver.HTTPServer(
            self.web_app, xheaders=self.trust_xheaders, ssl_options=ssl_options
        )

        for port in random_ports(self.port, self.port_retries + 1):
            try:
                self.http_server.listen(port, self.ip)
            except OSError as e:
                if e.errno == errno.EADDRINUSE:
                    self.log.info("The port %i is already in use, trying another port." % port)
                    continue
                elif e.errno in (errno.EACCES, getattr(errno, "WSAEACCES", errno.EACCES)):
                    self.log.warning("Permission to listen on port %i denied" % port)
                    continue
                else:
                    raise
            else:
                self.port = port
                break
        else:
            self.log.critical(
                "ERROR: the notebook server could not be started because "
                "no available port could be found."
            )
            self.exit(1)

    def init_io_loop(self):
        """init self.io_loop so that an extension can use it by io_loop.call_later() to create background tasks"""
        self.io_loop = ioloop.IOLoop.current()

    def init_signal(self):
        """Initialize signal handlers."""
        if not sys.platform.startswith("win") and sys.stdin and sys.stdin.isatty():
            signal.signal(signal.SIGINT, self._handle_sigint)
        signal.signal(signal.SIGTERM, self._signal_stop)
        if not sys.platform.startswith("win"):
            signal.signal(signal.SIGQUIT, self._signal_stop)

    def _handle_sigint(self, sig, frame):
        """SIGINT handler spawns confirmation dialog"""
        # register more forceful signal handler for ^C^C case
        signal.signal(signal.SIGINT, self._signal_stop)
        # request confirmation dialog in bg thread, to avoid
        # blocking the App
        thread = threading.Thread(target=self._confirm_exit)
        thread.daemon = True
        thread.start()

    def _restore_sigint_handler(self):
        """callback for restoring original SIGINT handler"""
        signal.signal(signal.SIGINT, self._handle_sigint)

    def _confirm_exit(self):
        """confirm shutdown on ^C

        A second ^C, or answering 'y' within 5s will cause shutdown,
        otherwise original SIGINT handler will be restored.

        This doesn't work on Windows.
        """
        info = self.log.info
        info("interrupted")
        # Check if answer_yes is set
        if self.answer_yes:
            self.log.critical("Shutting down...")
            # schedule stop on the main thread,
            # since this might be called from a signal handler
            self.stop(from_signal=True)
            return
        yes = "y"
        no = "n"
        sys.stdout.write("Shutdown this Jupyter server (%s/[%s])? " % (yes, no))  # noqa: UP031
        sys.stdout.flush()
        r, w, x = select.select([sys.stdin], [], [], 5)
        if r:
            line = sys.stdin.readline()
            if line.lower().startswith(yes) and no not in line.lower():
                self.log.critical("Shutdown confirmed")
                # schedule stop on the main thread,
                # since this might be called from a signal handler
                self.stop(from_signal=True)
                return
        else:
            info("No answer for 5s:")
        info("resuming operation...")
        # no answer, or answer is no:
        # set it back to original SIGINT handler
        # use IOLoop.add_callback because signal.signal must be called
        # from main thread
        self.io_loop.add_callback_from_signal(self._restore_sigint_handler)

    def _signal_stop(self, sig, frame):
        """Handle a stop signal."""
        self.log.critical("received signal %s, stopping", sig)
        self.stop(from_signal=True)

    def start_app(self):
        """Starts the application (with ioloop to follow)."""
        super().start()
        self.log.info(
            "Jupyter Kernel Gateway {} is available at http{}://{}:{}".format(
                KernelGatewayApp.version, "s" if self.keyfile else "", self.ip, self.port
            )
        )

    def start(self):
        """Starts an IO loop for the application."""

        self.start_app()

        if sys.platform != "win32":
            signal.signal(signal.SIGHUP, signal.SIG_IGN)

        signal.signal(signal.SIGTERM, self._signal_stop)

        try:
            self.io_loop.start()
        except KeyboardInterrupt:
            self.log.info("Interrupted...")
        finally:
            self.stop()

    async def _stop(self):
        """Cleanup resources and stop the IO Loop."""
        await self.personality.shutdown()
        await self.kernel_websocket_connection_class.close_all()
        if getattr(self, "io_loop", None):
            self.io_loop.stop()

    def stop(self, from_signal=False):
        """Cleanup resources and stop the server."""
        if hasattr(self, "http_server"):
            # Stop a server if its set.
            self.http_server.stop()
        if getattr(self, "io_loop", None):
            # use IOLoop.add_callback because signal.signal must be called
            # from main thread
            if from_signal:
                self.io_loop.add_callback_from_signal(self._stop)
            else:
                self.io_loop.add_callback(self._stop)

    def shutdown(self):
        """Stop all kernels in the pool."""
        self.io_loop.add_callback(self._stop)

    async def async_shutdown(self):
        """Stop all kernels in the pool."""
        if hasattr(self, "personality"):
            await self.personality.shutdown()


launch_instance = KernelGatewayApp.launch_instance



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Gateway Identity Provider interface

This defines the _authentication_ layer of Jupyter Server,
to be used in combination with Authorizer for _authorization_.
"""
from jupyter_server.auth.identity import IdentityProvider, User
from jupyter_server.base.handlers import JupyterHandler
from tornado import web
from traitlets import default


class GatewayIdentityProvider(IdentityProvider):
    """
    Interface for providing identity management and authentication for a Gateway server.
    """

    @default("token")
    def _token_default(self):
        return self.parent.auth_token

    @property
    def auth_enabled(self):
        if not self.token:
            return False
        return True

    def should_check_origin(self, handler: JupyterHandler) -> bool:
        """Should the Handler check for CORS origin validation?

        Origin check should be skipped for token-authenticated requests.

        Returns:
        - True, if Handler must check for valid CORS origin.
        - False, if Handler should skip origin check since requests are token-authenticated.
        """
        # Always check the origin unless operator configured gateway to allow any
        return handler.settings["kg_allow_origin"] != "*"

    def generate_anonymous_user(self, handler: web.RequestHandler) -> User:
        """Generate a random anonymous user.

        For use when a single shared token is used,
        but does not identify a user.
        """
        name = display_name = "Anonymous"
        initials = "An"
        color = None
        return User(name.lower(), name, display_name, initials, None, color)

    def is_token_authenticated(self, handler: web.RequestHandler) -> bool:
        """The default authentication flow of Gateway is token auth.

        The only other option is no auth
        """
        return True



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tornado handlers for kernel specs."""

import jupyter_server.kernelspecs.handlers as server_kernelspecs_resources_handlers
import jupyter_server.services.kernelspecs.handlers as server_handlers

from ...mixins import CORSMixin, JSONErrorsMixin, TokenAuthorizationMixin

# Extends the default handlers from the jupyter_server package with token auth, CORS
# and JSON errors.
default_handlers = []
for path, cls in server_handlers.default_handlers:
    # Everything should have CORS and token auth
    bases = (TokenAuthorizationMixin, CORSMixin, JSONErrorsMixin, cls)
    default_handlers.append((path, type(cls.__name__, bases, {})))

for path, cls in server_kernelspecs_resources_handlers.default_handlers:
    # Everything should have CORS and token auth
    bases = (TokenAuthorizationMixin, CORSMixin, JSONErrorsMixin, cls)
    default_handlers.append((path, type(cls.__name__, bases, {})))



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tests for swagger spec generation."""

import json

from kernel_gateway.notebook_http.cell.parser import APICellParser
from kernel_gateway.notebook_http.swagger.builders import SwaggerSpecBuilder
from kernel_gateway.notebook_http.swagger.parser import SwaggerCellParser


class TestSwaggerBuilders:
    """Unit tests the swagger spec builder."""

    def test_add_title_adds_title_to_spec(self):
        """Builder should store an API title."""
        expected = "Some New Title"
        builder = SwaggerSpecBuilder(APICellParser(comment_prefix="#"))
        builder.set_default_title(expected)
        result = builder.build()
        assert result["info"]["title"] == expected, "Title was not set to new value"

    def test_add_cell_adds_api_cell_to_spec(self):
        """Builder should store an API cell annotation."""
        expected = {"get": {"responses": {200: {"description": "Success"}}}}
        builder = SwaggerSpecBuilder(APICellParser(comment_prefix="#"))
        builder.add_cell("# GET /some/resource")
        result = builder.build()
        assert result["paths"]["/some/resource"] == expected, "Title was not set to new value"

    def test_all_swagger_preserved_in_spec(self):
        """Builder should store the swagger documented cell."""
        expected = """
        {
            "swagger": "2.0",
            "info" : {"version" : "0.0.0", "title" : "Default Title"},
            "paths": {
                "/some/resource": {
                    "get": {
                        "summary": "Get some resource",
                        "description": "Get some kind of resource?",
                        "operationId": "getSomeResource",
                        "produces": [
                            "application/json"
                        ],
                        "responses": {
                            "200": {
                                "description": "a resource",
                                "schema": {
                                    "type": "object",
                                    "required": ["name"],
                                    "properties": {
                                        "name": {
                                            "type": "string"
                                        }
                                    }
                                }
                            },
                            "400": {
                                "description": "Error retrieving resources",
                                "schema": {
                                    "$ref": "#/definitions/error"
                                }
                            }
                        }
                    }
                }
            }
        }
        """
        builder = SwaggerSpecBuilder(
            SwaggerCellParser(comment_prefix="#", notebook_cells=[{"source": expected}])
        )
        builder.add_cell(expected)
        result = builder.build()
        self.maxDiff = None
        assert (
            result["paths"]["/some/resource"]["get"]["description"]
            == json.loads(expected)["paths"]["/some/resource"]["get"]["description"]
        ), "description was not preserved"
        assert "info" in result, "info was not preserved"
        assert "title" in result["info"], "title was not present"
        assert (
            result["info"]["title"] == json.loads(expected)["info"]["title"]
        ), "title was not preserved"
        assert json.dumps(result["paths"]["/some/resource"], sort_keys=True) == json.dumps(
            json.loads(expected)["paths"]["/some/resource"], sort_keys=True
        ), "operations were not as expected"

        new_title = "new title. same contents."
        builder.set_default_title(new_title)
        result = builder.build()
        assert result["info"]["title"] != new_title, "title should not have been changed"

    def test_add_undocumented_cell_does_not_add_non_api_cell_to_spec(self):
        """Builder should store ignore non-API cells."""
        builder = SwaggerSpecBuilder(SwaggerCellParser(comment_prefix="#", notebook_cells=[]))
        builder.add_cell("regular code cell")
        builder.add_cell("# regular commented cell")
        result = builder.build()
        assert "paths" not in result, "unexpected paths were found"



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tests for notebook request utilities."""

import json
import unittest

from kernel_gateway.notebook_http.request_utils import (
    format_request,
    headers_to_dict,
    parameterize_path,
    parse_args,
    parse_body,
)


class MockRequest(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


class MockHeaders:
    def __init__(self, headers, **kwargs):
        self.headers = headers

    def get_all(self):
        return self.headers


class TestRequestUtils(unittest.TestCase):
    """Unit tests the request utility helper functions."""

    def test_parse_body_text(self):
        """Should parse the body text from a byte stream to a string."""
        request = MockRequest()
        request.body = b"test value"
        request.headers = {"Content-Type": "text/plain"}
        result = parse_body(request)
        self.assertEqual(result, "test value", "Did not properly parse text body.")

    def test_parse_body_json(self):
        """Should parse the body from a JSON byte stream to a dict."""
        request = MockRequest()
        request.body = b'{ "foo" : "bar" }'
        request.headers = {"Content-Type": "application/json"}
        result = parse_body(request)
        self.assertEqual(result, {"foo": "bar"}, "Did not properly parse json body.")

    def test_parse_body_bad_json(self):
        """Should parse the body from an invalid JSON byte stream to a string."""
        request = MockRequest()
        request.body = b'{ "foo" "bar" }'
        request.headers = {"Content-Type": "application/json"}
        result = parse_body(request)
        self.assertEqual(result, '{ "foo" "bar" }', "Did not properly parse json body.")

    def test_parse_body_multipart_form(self):
        """Should parse body arguments from multipart form data to a dict."""
        request = MockRequest()
        request.body = None
        request.body_arguments = {"foo": [b"bar"]}
        request.headers = {"Content-Type": "multipart/form-data"}
        result = parse_body(request)
        self.assertEqual(result, {"foo": ["bar"]}, "Did not properly parse json body.")

    def test_parse_body_url_encoded_form(self):
        """Should parse body arguments from urlencoded form data to a dict."""
        request = MockRequest()
        request.body = None
        request.body_arguments = {"foo": [b"bar"]}
        request.headers = {"Content-Type": "application/x-www-form-urlencoded"}
        result = parse_body(request)
        self.assertEqual(result, {"foo": ["bar"]}, "Did not properly parse json body.")

    def test_parse_body_empty(self):
        """Should parse an empty body to an empty string."""
        request = MockRequest()
        request.body = b""
        request.headers = {}
        result = parse_body(request)
        self.assertEqual(result, "", "Did not properly handle body = empty string.")

    def test_parse_body_defaults_to_text_plain(self):
        """Should parse a body to a string by default."""
        request = MockRequest()
        request.body = b'{"foo" : "bar"}'
        request.headers = {}
        result = parse_body(request)
        self.assertEqual(result, '{"foo" : "bar"}', "Did not properly handle body = empty string.")

    def test_parse_args(self):
        """Should parse URL argument byte streams to strings."""
        result = parse_args({"arga": [b"1234", b"4566"], "argb": [b"hello"]})
        self.assertEqual(
            result,
            {"arga": ["1234", "4566"], "argb": ["hello"]},
            "Did not properly convert query parameters.",
        )

    def test_parameterize_path(self):
        """Should parse URLs with path parameters into regular expressions."""
        result = parameterize_path("/foo/:bar")
        self.assertEqual(result, r"/foo/(?P<bar>[^\/]+)")
        result = parameterize_path("/foo/:bar/baz/:quo")
        self.assertEqual(result, r"/foo/(?P<bar>[^\/]+)/baz/(?P<quo>[^\/]+)")

    def test_whitespace_in_paths(self):
        """Should handle whitespace in the path."""
        result = parameterize_path("/foo/:bar ")
        self.assertEqual(result, r"/foo/(?P<bar>[^\/]+)")
        result = parameterize_path("/foo/:bar/baz ")
        self.assertEqual(result, r"/foo/(?P<bar>[^\/]+)/baz")

    def test_headers_to_dict(self):
        """Should parse headers into a dictionary."""
        result = headers_to_dict(
            MockHeaders(
                [("Content-Type", "application/json"), ("Set-Cookie", "A=B"), ("Set-Cookie", "C=D")]
            )
        )
        self.assertEqual(
            result["Content-Type"],
            "application/json",
            "Single value for header was not assigned correctly",
        )
        self.assertEqual(
            result["Set-Cookie"],
            ["A=B", "C=D"],
            "Single value for header was not assigned correctly",
        )

    def test_headers_to_dict_with_no_headers(self):
        """Should parse empty headers into an empty dictionary."""
        result = headers_to_dict(MockHeaders([]))
        self.assertEqual(
            result, {}, "Empty headers handled incorrectly and did not make empty dict"
        )

    def test_format_request_code_not_escaped(self):
        """Should handle quotes in headers."""
        test_request = """{"body": "", "headers": {"Accept-Language": "en-US,en;q=0.8",
                        "If-None-Match": "9a28a9262f954494a8de7442c63d6d0715ce0998",
                        "Accept-Encoding": "gzip, deflate, sdch"}, "args": {}, "path": {}}"""
        request_code = format_request(test_request)
        # Get the value of REQUEST = "{ to test for equality
        test_request_js_value = request_code[request_code.index('"{') :]
        self.assertEqual(
            test_request,
            json.loads(test_request_js_value),
            "Request code without escaped quotes was not formatted correctly",
        )

    def test_format_request_code_escaped(self):
        """Should handle backslash escaped characters in headers."""
        test_request = """{"body": "", "headers": {"Accept-Language": "en-US,en;q=0.8",
                        "If-None-Match": "\"\"9a28a9262f954494a8de7442c63d6d0715ce0998\"\"",
                        "Accept-Encoding": "gzip, deflate, sdch"}, "args": {}, "path": {}}"""
        request_code = format_request(test_request)
        # Get the value of REQUEST = "{ to test for equality
        test_request_js_value = request_code[request_code.index('"{') :]
        self.assertEqual(
            test_request,
            json.loads(test_request_js_value),
            "Escaped Request code was not formatted correctly",
        )

    def test_format_request_without_a_kernel_language_arg(self):
        test_request = """{"body": "", "headers": {}, "args": {}, "path": {}}"""
        request_code = format_request(test_request)
        self.assertTrue(
            request_code.startswith("REQUEST"),
            "Call format_request without a kernel_language argument was not formatted correctly",
        )

    def test_format_request_with_a_kernel_language_python(self):
        test_request = """{"body": "", "headers": {}, "args": {}, "path": {}}"""
        request_code = format_request(test_request, "python")
        self.assertTrue(
            request_code.startswith("REQUEST"),
            'Call format_request with a kernel_language "python" was not formatted correctly',
        )

    def test_format_request_with_a_kernel_language_scala(self):
        test_request = """{"body": "", "headers": {}, "args": {}, "path": {}}"""
        request_code = format_request(test_request, "scala")
        self.assertTrue(
            request_code.startswith("REQUEST"),
            'Call format_request with a kernel_language "scala" was not formatted correctly',
        )

    def test_format_request_with_a_kernel_language_perl(self):
        test_request = """{"body": "", "headers": {}, "args": {}, "path": {}}"""
        request_code = format_request(test_request, "perl")
        self.assertTrue(
            request_code.startswith("my $REQUEST"),
            'Call format_request with a kernel language "perl" was not formatted correctly',
        )

    def test_format_request_with_a_kernel_language_bash(self):
        test_request = """{"body": "", "headers": {}, "args": {}, "path": {}}"""
        request_code = format_request(test_request, "bash")
        self.assertTrue(
            request_code.startswith('REQUEST="{'),
            'Call format_request with a kernel language "bash" was not formatted correctly',
        )



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tests for notebook cell parsing."""

from kernel_gateway.notebook_http.swagger.parser import SwaggerCellParser


class TestSwaggerAPICellParser:
    """Unit tests the SwaggerCellParser class."""

    def test_basic_swagger_parse(self):
        """Parser should correctly identify Swagger cells."""
        parser = SwaggerCellParser(
            comment_prefix="#",
            notebook_cells=[
                {
                    "source": '```\n{"swagger":"2.0", "paths": {"": {"post": {"operationId": "foo", "parameters": [{"name": "foo"}]}}}}\n```\n'
                }
            ],
        )
        assert "swagger" in parser.swagger, "Swagger doc was not detected"

    def test_basic_is_api_cell(self):
        """Parser should correctly identify operation cells."""
        parser = SwaggerCellParser(
            comment_prefix="#",
            notebook_cells=[
                {
                    "source": '```\n{"swagger":"2.0", "paths": {"": {"post": {"operationId": "foo", "parameters": [{"name": "foo"}]}}}}\n```\n'
                }
            ],
        )
        assert parser.is_api_cell("#operationId:foo"), "API cell was not detected with " + str(
            parser.kernelspec_operation_indicator
        )
        assert parser.is_api_cell("# operationId:foo"), "API cell was not detected with " + str(
            parser.kernelspec_operation_indicator
        )
        assert parser.is_api_cell("#operationId: foo"), "API cell was not detected with " + str(
            parser.kernelspec_operation_indicator
        )
        assert parser.is_api_cell("no") is False, "API cell was detected"
        assert parser.is_api_cell("# another comment") is False, "API cell was detected"

    def test_basic_is_api_response_cell(self):
        """Parser should correctly identify ResponseInfo cells."""
        parser = SwaggerCellParser(
            comment_prefix="#",
            notebook_cells=[
                {
                    "source": '```\n{"swagger":"2.0", "paths": {"": {"post": {"operationId": "foo", "parameters": [{"name": "foo"}]}}}}\n```\n'
                }
            ],
        )
        assert parser.is_api_response_cell("#ResponseInfo operationId:foo"), (
            "Response cell was not detected with "
            + str(parser.kernelspec_operation_response_indicator)
        )
        assert parser.is_api_response_cell("# ResponseInfo operationId:foo"), (
            "Response cell was not detected with "
            + str(parser.kernelspec_operation_response_indicator)
        )
        assert parser.is_api_response_cell("# ResponseInfo  operationId: foo"), (
            "Response cell was not detected with "
            + str(parser.kernelspec_operation_response_indicator)
        )
        assert parser.is_api_response_cell("#ResponseInfo operationId: foo"), (
            "Response cell was not detected with "
            + str(parser.kernelspec_operation_response_indicator)
        )
        assert parser.is_api_response_cell("# operationId: foo") is False, (
            "API cell was detected as a ResponseInfo cell "
            + str(parser.kernelspec_operation_response_indicator)
        )
        assert parser.is_api_response_cell("no") is False, "API cell was detected"

    def test_endpoint_sort_default_strategy(self):
        """Parser should sort duplicate endpoint paths."""
        source_cells = [
            {
                "source": '\n```\n{"swagger":"2.0","paths":{"":{"post":{"operationId":"postRoot","parameters":[{"name":"foo"}]}},"/hello":{"post":{"operationId":"postHello","parameters":[{"name":"foo"}]},"get":{"operationId":"getHello","parameters":[{"name":"foo"}]}},"/hello/world":{"put":{"operationId":"putWorld"}}}}\n```\n'
            },
            {"source": "# operationId:putWorld"},
            {"source": "# operationId:getHello"},
            {"source": "# operationId:postHello"},
            {"source": "# operationId:postRoot"},
        ]
        parser = SwaggerCellParser(comment_prefix="#", notebook_cells=source_cells)
        endpoints = parser.endpoints(cell["source"] for cell in source_cells)

        expected_values = ["/hello/world", "/hello/:foo", "/:foo"]
        try:
            for index in range(len(expected_values)):
                endpoint, _ = endpoints[index]
                assert (
                    expected_values[index] == endpoint
                ), "Endpoint was not found in expected order"
        except IndexError:
            raise RuntimeError(endpoints) from None

    def test_endpoint_sort_custom_strategy(self):
        """Parser should sort duplicate endpoint paths using a custom sort strategy."""
        source_cells = [
            {
                "source": '```\n{"swagger": "2.0", "paths": {"/1": {"post": {"operationId": "post1"}},"/+": {"post": {"operationId": "postPlus"}},"/a": {"get": {"operationId": "getA"}}}}\n```\n'
            },
            {"source": "# operationId: post1"},
            {"source": "# operationId: postPlus"},
            {"source": "# operationId: getA"},
        ]

        def custom_sort_fun(endpoint):
            if endpoint.find("1") >= 0:
                return 0
            elif endpoint.find("a") >= 0:
                return 1
            else:
                return 2

        parser = SwaggerCellParser(comment_prefix="#", notebook_cells=source_cells)
        endpoints = parser.endpoints((cell["source"] for cell in source_cells), custom_sort_fun)
        print(str(endpoints))

        expected_values = ["/+", "/a", "/1"]
        for index in range(len(expected_values)):
            endpoint, _ = endpoints[index]
            assert expected_values[index] == endpoint, "Endpoint was not found in expected order"

    def test_get_cell_endpoint_and_verb(self):
        """Parser should extract API endpoint and verb from cell annotations."""
        parser = SwaggerCellParser(
            comment_prefix="#",
            notebook_cells=[
                {
                    "source": '```\n{"swagger":"2.0", "paths": {"/foo": {"get": {"operationId": "getFoo"}}, "/bar/quo": {"post": {"operationId": "post_bar_Quo"}}}}\n```\n'
                }
            ],
        )
        endpoint, verb = parser.get_cell_endpoint_and_verb("# operationId: getFoo")
        assert endpoint == "/foo", "Endpoint was not extracted correctly"
        assert verb.lower() == "get", "Endpoint was not extracted correctly"
        endpoint, verb = parser.get_cell_endpoint_and_verb("# operationId: post_bar_Quo")
        assert endpoint == "/bar/quo", "Endpoint was not extracted correctly"
        assert verb.lower() == "post", "Endpoint was not extracted correctly"

        endpoint, verb = parser.get_cell_endpoint_and_verb("some regular code")
        assert (
            endpoint is None
        ), "Endpoint was not extracted correctly (something was actually returned)"
        assert (
            verb is None
        ), "Endpoint was not extracted correctly (something was actually returned)"

    def test_endpoint_concatenation(self):
        """Parser should concatenate multiple cells with the same verb+path."""
        cells = [
            {
                "source": '```\n{"swagger":"2.0", "paths": {"/foo": {"put": {"operationId":"putFoo","parameters": [{"name": "bar"}]},"post":{"operationId":"postFooBody"},"get": {"operationId":"getFoo","parameters": [{"name": "bar"}]}}}}\n```\n'
            },
            {"source": "# operationId: postFooBody "},
            {"source": "# unrelated comment "},
            {"source": "# operationId: putFoo"},
            {"source": "# operationId: puttFoo"},
            {"source": "# operationId: getFoo"},
            {"source": "# operationId: putFoo"},
        ]
        parser = SwaggerCellParser(comment_prefix="#", notebook_cells=cells)
        endpoints = parser.endpoints(cell["source"] for cell in cells)
        assert len(endpoints) == 2, endpoints
        # for ease of testing
        endpoints = dict(endpoints)
        assert len(endpoints["/foo"]) == 1
        assert len(endpoints["/foo/:bar"]) == 2
        assert endpoints["/foo"]["post"] == "# operationId: postFooBody \n"
        assert endpoints["/foo/:bar"]["get"] == "# operationId: getFoo\n"
        assert endpoints["/foo/:bar"]["put"] == "# operationId: putFoo\n# operationId: putFoo\n"

    def test_endpoint_response_concatenation(self):
        """Parser should concatenate multiple response cells with the same verb+path."""
        source_cells = [
            {
                "source": '```\n{"swagger":"2.0", "paths": {"/foo": {"put": {"operationId":"putbar","parameters": [{"name": "bar"}]},"post":{"operationId":"postbar"},"get": {"operationId":"get","parameters": [{"name": "bar"}]}}}}\n```\n'
            },
            {"source": "# ResponseInfo operationId: get"},
            {"source": "# ResponseInfo operationId: postbar "},
            {"source": "# ResponseInfo operationId: putbar"},
            {"source": "# ResponseInfo operationId: puttbar"},
            {"source": "ignored"},
            {"source": "# ResponseInfo operationId: putbar "},
        ]
        parser = SwaggerCellParser(comment_prefix="#", notebook_cells=source_cells)
        endpoints = parser.endpoint_responses(cell["source"] for cell in source_cells)
        assert len(endpoints) == 2
        # for ease of testing
        endpoints = dict(endpoints)
        assert len(endpoints["/foo"]) == 1
        assert len(endpoints["/foo/:bar"]) == 2
        assert endpoints["/foo"]["post"] == "# ResponseInfo operationId: postbar \n"
        assert (
            endpoints["/foo/:bar"]["put"]
            == "# ResponseInfo operationId: putbar\n# ResponseInfo operationId: putbar \n"
        )
        assert endpoints["/foo/:bar"]["get"] == "# ResponseInfo operationId: get\n"

    def test_undeclared_operations(self, caplog):
        """Parser should warn about operations that aren't documented in the swagger cell."""
        source_cells = [
            {
                "source": '```\n{"swagger":"2.0", "paths": {"/foo": {"put": {"operationId":"putbar","parameters": [{"name": "bar"}]},"post":{"operationId":"postbar"},"get": {"operationId":"get","parameters": [{"name": "bar"}]}}}}\n```\n'
            },
            {"source": "# operationId: get"},
            {"source": "# operationId: postbar "},
            {"source": "# operationId: putbar"},
            {"source": "# operationId: extraOperation"},
        ]

        SwaggerCellParser(comment_prefix="#", notebook_cells=source_cells)
        assert "extraOperation" in caplog.text

    def test_undeclared_operations_reversed(self, caplog):
        """Parser should warn about operations that aren"t documented in the swagger cell."""
        source_cells = [
            {"source": "# operationId: get"},
            {"source": "# operationId: postbar "},
            {"source": "# operationId: putbar"},
            {"source": "# operationId: extraOperation"},
            {
                "source": '```\n{"swagger":"2.0", "paths": {"/foo": {"put": {"operationId":"putbar","parameters": [{"name": "bar"}]},"post":{"operationId":"postbar"},"get": {"operationId":"get","parameters": [{"name": "bar"}]}}}}\n```\n'
            },
        ]

        SwaggerCellParser(comment_prefix="#", notebook_cells=source_cells)
        assert "extraOperation" in caplog.text

    def test_unreferenced_operations(self, caplog):
        """Parser should warn about documented operations that aren"t referenced in a cell."""
        source_cells = [
            {
                "source": '```\n{"swagger":"2.0", "paths": {"/foo": {"put": {"operationId":"putbar","parameters": [{"name": "bar"}]},"post":{"operationId":"postbar"},"get": {"operationId":"get","parameters": [{"name": "bar"}]}}}}\n```\n'
            },
            {"source": "# operationId: get"},
            {"source": "# operationId: putbar"},
            {"source": "# operationId: putbar "},
        ]

        SwaggerCellParser(comment_prefix="#", notebook_cells=source_cells)
        assert "postbar" in caplog.text



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Notebook HTTP personality for the Kernel Gateway"""

import importlib
import os

import tornado
from jupyter_server.utils import url_path_join
from traitlets import Bool, Dict, Unicode, default
from traitlets.config.configurable import LoggingConfigurable

from ..base.handlers import default_handlers as default_base_handlers
from ..services.kernels.pool import ManagedKernelPool
from .cell.parser import APICellParser  # noqa: F401
from .handlers import NotebookAPIHandler, NotebookDownloadHandler, parameterize_path
from .swagger.handlers import SwaggerSpecHandler


class NotebookHTTPPersonality(LoggingConfigurable):
    """Personality for notebook-http support, creating REST endpoints
    based on the notebook's annotated cells
    """

    cell_parser_env = "KG_CELL_PARSER"
    cell_parser = Unicode(
        "kernel_gateway.notebook_http.cell.parser",
        config=True,
        help="""Determines which module is used to parse the notebook for endpoints and
            documentation. Valid module names include 'kernel_gateway.notebook_http.cell.parser'
            and 'kernel_gateway.notebook_http.swagger.parser'. (KG_CELL_PARSER env var)
            """,
    )

    @default("cell_parser")
    def cell_parser_default(self):
        return os.getenv(self.cell_parser_env, "kernel_gateway.notebook_http.cell.parser")

    # Intentionally not defining an env var option for a dict type
    comment_prefix = Dict(
        {"scala": "//", None: "#"}, config=True, help="Maps kernel language to code comment syntax"
    )

    allow_notebook_download_env = "KG_ALLOW_NOTEBOOK_DOWNLOAD"
    allow_notebook_download = Bool(
        config=True,
        help="Optional API to download the notebook source code in notebook-http mode, defaults to not allow",
    )

    @default("allow_notebook_download")
    def allow_notebook_download_default(self):
        return os.getenv(self.allow_notebook_download_env, "False") == "True"

    static_path_env = "KG_STATIC_PATH"
    static_path = Unicode(
        None,
        config=True,
        allow_none=True,
        help="Serve static files on disk in the given path as /public, defaults to not serve",
    )

    @default("static_path")
    def static_path_default(self):
        return os.getenv(self.static_path_env)

    kernel_pool: ManagedKernelPool

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        # Import the module to use for cell endpoint parsing
        cell_parser_module = importlib.import_module(self.cell_parser)
        # Build the parser using the comment syntax for the notebook language
        func = cell_parser_module.create_parser
        try:
            kernel_language = self.parent.seed_notebook["metadata"]["language_info"]["name"]
        except (AttributeError, KeyError):
            kernel_language = None
        prefix = self.comment_prefix.get(kernel_language, "#")
        self.api_parser = func(
            parent=self,
            log=self.log,
            comment_prefix=prefix,
            notebook_cells=self.parent.seed_notebook.cells,
        )
        self.kernel_language = kernel_language
        self.kernel_pool = ManagedKernelPool()

    async def init_configurables(self):
        await self.kernel_pool.initialize(self.parent.prespawn_count, self.parent.kernel_manager)

    def create_request_handlers(self):
        """Create handlers and redefine them off of the base_url path. Assumes
        init_configurables() has already been called, and that the seed source
        was available there.
        """
        handlers = []
        # Register the NotebookDownloadHandler if configuration allows
        if self.allow_notebook_download:
            path = url_path_join("/", self.parent.base_url, r"/_api/source")
            self.log.info(f"Registering resource: {path}, methods: (GET)")
            handlers.append((path, NotebookDownloadHandler, {"path": self.parent.seed_uri}))

        # Register a static path handler if configuration allows
        if self.static_path is not None:
            path = url_path_join("/", self.parent.base_url, r"/public/(.*)")
            self.log.info(f"Registering resource: {path}, methods: (GET)")
            handlers.append((path, tornado.web.StaticFileHandler, {"path": self.static_path}))

        # Discover the notebook endpoints and their implementations
        endpoints = self.api_parser.endpoints(self.parent.kernel_manager.seed_source)
        response_sources = self.api_parser.endpoint_responses(
            self.parent.kernel_manager.seed_source
        )
        if len(endpoints) == 0:
            raise RuntimeError(
                "No endpoints were discovered. Check your notebook to make sure your cells are annotated correctly."
            )

        # Cycle through the (endpoint_path, source) tuples and register their handlers
        for endpoint_path, verb_source_map in endpoints:
            parameterized_path = parameterize_path(endpoint_path)
            parameterized_path = url_path_join("/", self.parent.base_url, parameterized_path)
            self.log.info(
                "Registering resource: {}, methods: ({})".format(
                    parameterized_path, list(verb_source_map.keys())
                )
            )
            response_source_map = (
                response_sources[endpoint_path] if endpoint_path in response_sources else {}
            )
            handler_args = {
                "sources": verb_source_map,
                "response_sources": response_source_map,
                "kernel_pool": self.kernel_pool,
                "kernel_name": self.parent.kernel_manager.seed_kernelspec,
                "kernel_language": self.kernel_language or "",
            }
            handlers.append((parameterized_path, NotebookAPIHandler, handler_args))

        # Register the swagger API spec handler
        path = url_path_join("/", self.parent.base_url, r"/_api/spec/swagger.json")
        handlers.append(
            (
                path,
                SwaggerSpecHandler,
                {
                    "notebook_path": self.parent.seed_uri,
                    "source_cells": self.parent.seed_notebook.cells,
                    "cell_parser": self.api_parser,
                },
            )
        )
        self.log.info(f"Registering resource: {path}, methods: (GET)")

        # Add the 404 catch-all last
        handlers.append(default_base_handlers[-1])
        return handlers

    def should_seed_cell(self, code):
        """Determines whether the given code cell source should be executed when
        seeding a new kernel."""
        # seed cells that are uninvolved with the presented API
        return not self.api_parser.is_api_cell(code) and not self.api_parser.is_api_response_cell(
            code
        )

    async def shutdown(self):
        """Stop all kernels in the pool."""
        await self.kernel_pool.shutdown()


def create_personality(*args, **kwargs):
    return NotebookHTTPPersonality(*args, **kwargs)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tests for notebook cell parsing."""

import sys

from kernel_gateway.notebook_http.cell.parser import APICellParser


class TestAPICellParser:
    """Unit tests the APICellParser class."""

    def test_is_api_cell(self):
        """Parser should correctly identify annotated API cells."""
        parser = APICellParser(comment_prefix="#")
        assert parser.is_api_cell("# GET /yes"), "API cell was not detected"
        assert parser.is_api_cell("no") is False, "API cell was not detected"

    def test_endpoint_sort_default_strategy(self):
        """Parser should sort duplicate endpoint paths."""
        source_cells = [
            "# POST /:foo",
            "# POST /hello/:foo",
            "# GET /hello/:foo",
            "# PUT /hello/world",
        ]
        parser = APICellParser(comment_prefix="#")
        endpoints = parser.endpoints(source_cells)
        expected_values = ["/hello/world", "/hello/:foo", "/:foo"]

        for index in range(len(expected_values)):
            endpoint, _ = endpoints[index]
            assert expected_values[index] == endpoint, "Endpoint was not found in expected order"

    def test_endpoint_sort_custom_strategy(self):
        """Parser should sort duplicate endpoint paths using a custom sort
        strategy.
        """
        source_cells = ["# POST /1", "# POST /+", "# GET /a"]

        def custom_sort_fun(endpoint):
            _ = sys.maxsize
            if endpoint.find("1") >= 0:
                return 0
            elif endpoint.find("a") >= 0:
                return 1
            else:
                return 2

        parser = APICellParser(comment_prefix="#")
        endpoints = parser.endpoints(source_cells, custom_sort_fun)
        expected_values = ["/+", "/a", "/1"]

        for index in range(len(expected_values)):
            endpoint, _ = endpoints[index]
            assert expected_values[index] == endpoint, "Endpoint was not found in expected order"

    def test_get_cell_endpoint_and_verb(self):
        """Parser should extract API endpoint and verb from cell annotations."""
        parser = APICellParser(comment_prefix="#")
        endpoint, verb = parser.get_cell_endpoint_and_verb("# GET /foo")
        assert endpoint, "/foo" == "Endpoint was not extracted correctly"
        assert verb, "GET" == "Endpoint was not extracted correctly"
        endpoint, verb = parser.get_cell_endpoint_and_verb("# POST /bar/quo")
        assert endpoint, "/bar/quo" == "Endpoint was not extracted correctly"
        assert verb, "POST" == "Endpoint was not extracted correctly"

        endpoint, verb = parser.get_cell_endpoint_and_verb("some regular code")
        assert endpoint is None, "Endpoint was not extracted correctly"
        assert verb is None, "Endpoint was not extracted correctly"

    def test_endpoint_concatenation(self):
        """Parser should concatenate multiple cells with the same verb+path."""
        source_cells = [
            "# POST /foo/:bar",
            "# POST /foo/:bar",
            "# POST /foo",
            "ignored",
            "# GET /foo/:bar",
        ]
        parser = APICellParser(comment_prefix="#")
        endpoints = parser.endpoints(source_cells)
        assert len(endpoints) == 2
        # for ease of testing
        endpoints = dict(endpoints)
        assert len(endpoints["/foo"]) == 1
        assert len(endpoints["/foo/:bar"]) == 2
        assert endpoints["/foo"]["POST"] == "# POST /foo\n"
        assert endpoints["/foo/:bar"]["POST"] == "# POST /foo/:bar\n# POST /foo/:bar\n"
        assert endpoints["/foo/:bar"]["GET"] == "# GET /foo/:bar\n"

    def test_endpoint_response_concatenation(self):
        """Parser should concatenate multiple response cells with the same
        verb+path.
        """
        source_cells = [
            "# ResponseInfo POST /foo/:bar",
            "# ResponseInfo POST /foo/:bar",
            "# ResponseInfo POST /foo",
            "ignored",
            "# ResponseInfo GET /foo/:bar",
        ]
        parser = APICellParser(comment_prefix="#")
        endpoints = parser.endpoint_responses(source_cells)
        assert len(endpoints) == 2
        # for ease of testing
        endpoints = dict(endpoints)
        assert len(endpoints["/foo"]) == 1
        assert len(endpoints["/foo/:bar"]) == 2
        assert endpoints["/foo"]["POST"] == "# ResponseInfo POST /foo\n"
        assert (
            endpoints["/foo/:bar"]["POST"]
            == "# ResponseInfo POST /foo/:bar\n# ResponseInfo POST /foo/:bar\n"
        )
        assert endpoints["/foo/:bar"]["GET"] == "# ResponseInfo GET /foo/:bar\n"



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tornado handler that returns the swagger specification of a
notebook-http mode, notebook-defined API.
"""

import json

import tornado.web

from ...mixins import CORSMixin, JSONErrorsMixin, TokenAuthorizationMixin
from .builders import SwaggerSpecBuilder


class SwaggerSpecHandler(
    TokenAuthorizationMixin, CORSMixin, JSONErrorsMixin, tornado.web.RequestHandler
):
    """Handles requests for the Swagger specification of a notebook defined
    API.
    """

    output = None

    def initialize(self, notebook_path, source_cells, cell_parser):
        """Builds the spec for the notebook-defined API.

        Parameters
        ----------
        notebook_path : str
            Path to the notebook, used to set the API title
        source_cells : list
            code cells
        cell_parser : obj
            Parser that will return endpoints and documentation
        """
        if self.output is None:
            spec_builder = SwaggerSpecBuilder(cell_parser)
            for source_cell in source_cells:
                if "source" in source_cell:
                    spec_builder.add_cell(source_cell["source"])
            spec_builder.set_default_title(notebook_path)
            SwaggerSpecHandler.output = json.dumps(spec_builder.build())

    async def get(self, **kwargs):
        """Responds with the spec in JSON format."""
        self.set_header("Content-Type", "application/json")
        self.set_status(200)
        await self.finish(self.output)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Entrypoint for the kernel gateway package."""
from ._version import __version__, version_info  # noqa: F401
from .gatewayapp import launch_instance  # noqa: F401



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tornado handlers for invoking notebook cells as web APIs."""

import json
import os
from functools import partial
from typing import Optional

import tornado.web
from tornado.concurrent import Future
from tornado.log import access_log

from ..mixins import CORSMixin, JSONErrorsMixin, TokenAuthorizationMixin
from .errors import CodeExecutionError, UnsupportedMethodError
from .request_utils import (
    format_request,
    headers_to_dict,
    parameterize_path,  # noqa: F401
    parse_args,
    parse_body,
)


class NotebookAPIHandler(
    TokenAuthorizationMixin, CORSMixin, JSONErrorsMixin, tornado.web.RequestHandler
):
    """Executes code from a notebook cell in response to HTTP requests at the
    route registered in association with this class.

    Supports the GET, POST, PUT, and DELETE HTTP methods.

    Parameters
    ----------
    sources : dict
        Maps HTTP verb strings to annotated cells extracted from a notebook
    response_sources
        Maps HTTP verb strings to ResponseInfo annotated cells extracted from a
        notebook
    kernel_pool
        Instance of services.kernels.ManagedKernelPool
    kernel_name
        Kernel spec name used to launch the kernel pool
    kernel_language
        Kernel spec language used to make language specific operations

    Attributes
    ----------
    See parameters: they are stored as passed as instance variables.

    See Also
    --------
    services.cell.parser.APICellParser for detail about how the source cells
    are identified, parsed, and associated with HTTP verbs and paths.
    """

    def initialize(self, sources, response_sources, kernel_pool, kernel_name, kernel_language=""):
        self.kernel_pool = kernel_pool
        self.sources = sources
        self.kernel_name = kernel_name
        self.response_sources = response_sources
        self.kernel_language = kernel_language

    def finish_future(self, future, result_accumulator):
        """Resolves the promise to respond to a HTTP request handled by a
        kernel in the pool.

        Defines precedence for the kind of response:

        1. If any error occurred, resolve with a CodeExecutionError exception
        2. If any stream message was collected from the kernel, resolve with
            the joined string of stream messages
        3. If an execute_result was collected, resolve with that result object
            JSON encoded
        4. Resolve with an empty string

        Parameters
        ----------
        future : tornado.concurrent.Future
            Promise of a future response to an API request
        result_accumulator : dict
            Dictionary of results from a kernel with at least the keys error,
            stream, and result
        """
        if result_accumulator["error"]:
            future.set_exception(CodeExecutionError(result_accumulator["error"]))
        elif len(result_accumulator["stream"]) > 0:
            future.set_result("".join(result_accumulator["stream"]))
        elif result_accumulator["result"]:
            future.set_result(json.dumps(result_accumulator["result"]))
        else:
            # If nothing was set, return an empty value
            future.set_result("")

    def on_recv(self, result_accumulator, future, parent_header, msg):
        """Collects ipoub messages associated with code execution request
        identified by `parent_header`.

        Continues collecting until an execution state of idle is reached.
        The first three parameters are typically applied in a partial.

        Parameters
        ----------
        result_accumulator : dict
            Accumulates data from `execute_result`, `stream` and `error`
            messages under keys `result`, `stream`, and `error` respectively
            across multiple invocations of this callback.
        future : tornado.concurrent.Future
            Promise to resolve when the kernel goes idle
        parent_header : dict
            Parent header from an `execute` request, used to identify messages
            that relate to its execution vs other executions
        msg : dict
            Kernel message received from the iopub channel
        """
        if msg["parent_header"]["msg_id"] == parent_header:
            # On idle status, exit our loop
            if (
                msg["header"]["msg_type"] == "status"
                and msg["content"]["execution_state"] == "idle"
            ):
                self.finish_future(future, result_accumulator)
            # Store the execute result
            elif msg["header"]["msg_type"] == "execute_result":
                result_accumulator["result"] = msg["content"]["data"]
            # Accumulate the stream messages
            elif msg["header"]["msg_type"] == "stream":
                # Only take stream output if it is on stdout or if the kernel
                # is non-confirming and does not name the stream
                if "name" not in msg["content"] or msg["content"]["name"] == "stdout":
                    result_accumulator["stream"].append(msg["content"]["text"])
            # Store the error message
            elif msg["header"]["msg_type"] == "error":
                error_name = msg["content"]["ename"]
                error_value = msg["content"]["evalue"]
                result_accumulator["error"] = f"Error {error_name}: {error_value} \n"

    def execute_code(self, kernel_client, kernel_id, source_code):
        """Executes `source_code` on the kernel specified.

        Registers a callback for iopub messages. Promises to return the output
        of the execution in the future after the kernel returns to its idle
        state.

        Parameters
        ----------
        kernel_client : object
            Client to use to execute the code
        kernel_id : str
            ID of the kernel from the pool that will execute the request
        source_code : str
            Source code to execute

        Returns
        -------
        tornado.concurrent.Future
            Promise of execution result

        Raises
        ------
        CodeExecutionError
            If the kernel returns any error
        """
        future = Future()
        result_accumulator = {"stream": [], "error": None, "result": None}
        parent_header = kernel_client.execute(source_code)
        on_recv_func = partial(self.on_recv, result_accumulator, future, parent_header)
        self.kernel_pool.on_recv(kernel_id, on_recv_func)
        return future

    async def _handle_request(self):
        """Turns an HTTP request into annotated notebook code to execute on a
        kernel.

        Sets the HTTP response code, headers, and response body based on the
        result of the kernel execution. Then finishes the Tornado response.
        """
        self.response_future = Future()
        kernel_client, kernel_id = await self.kernel_pool.acquire()
        try:
            # Method not supported
            if self.request.method not in self.sources:
                raise UnsupportedMethodError(self.request.method)

            # Set the Content-Type and status to default values
            self.set_header("Content-Type", "text/plain")
            self.set_status(200)

            # Get the source to execute in response to this request
            source_code = self.sources[self.request.method]
            # Build the request dictionary
            request = json.dumps(
                {
                    "body": parse_body(self.request),
                    "args": parse_args(self.request.query_arguments),
                    "path": self.path_kwargs,
                    "headers": headers_to_dict(self.request.headers),
                }
            )
            # Turn the request string into a valid code string
            request_code = format_request(request, self.kernel_language)

            # Run the request and source code and yield until there's a result
            access_log.debug(f"Request code for notebook cell is: {request_code}")
            await self.execute_code(kernel_client, kernel_id, request_code)
            source_result = await self.execute_code(kernel_client, kernel_id, source_code)

            # If a response code cell exists, execute it
            if self.request.method in self.response_sources:
                response_code = self.response_sources[self.request.method]
                response_future = self.execute_code(kernel_client, kernel_id, response_code)

                # Wait for the response and parse the json value
                response_result = await response_future
                response = json.loads(response_result)

                # Copy all the header values into the tornado response
                if "headers" in response:
                    for header in response["headers"]:
                        self.set_header(header, response["headers"][header])

                # Set the status code if it exists
                if "status" in response:
                    self.set_status(response["status"])

            # Write the result of the source code execution
            if source_result:
                self.write(source_result)
        # If there was a problem executing an code, return a 500
        except CodeExecutionError as err:
            self.write(str(err))
            self.set_status(500)
        # An unsupported method was called on this handler
        except UnsupportedMethodError:
            self.set_status(405)
        finally:
            # Always make sure we release the kernel and finish the request
            self.response_future.set_result(None)
            self.kernel_pool.release(kernel_id)
            await self.finish()

    async def get(self, **kwargs):
        await self._handle_request()
        await self.response_future

    async def post(self, **kwargs):
        await self._handle_request()
        await self.response_future

    async def put(self, **kwargs):
        await self._handle_request()
        await self.response_future

    async def delete(self, **kwargs):
        await self._handle_request()
        await self.response_future

    def options(self, **kwargs):
        self.finish()


class NotebookDownloadHandler(
    TokenAuthorizationMixin, CORSMixin, JSONErrorsMixin, tornado.web.StaticFileHandler
):
    """Handles requests to download the annotated notebook behind the web API."""

    def initialize(self, path: str, default_filename: Optional[str] = None):
        self.dirname, self.filename = os.path.split(path)
        super().initialize(self.dirname)

    async def get(self, include_body: bool = True):
        res = await super().get(path=self.filename, include_body=include_body)
        return res
